
		; Scripts for TeamTalk5Classic
;
; Author:  Doug Lee

/*
Design note:
These scripts use the Scripting.Dictionary object available in all
supported versions of Windows, rather than the Collection data type
included in the JAWS scripting language starting in JAWS 11 update 2,
so that they can compile and run under older JAWS versions. The usage
and syntax of Dictionary objects, compared to Collections, is a bit
slower and messier but is not problematic unless hundreds of
references to a Dictionary object occur in one second. Would-be
modifiers of these scripts should take care to minimize such
references in loops.
[DGL, 2011-12-18]
*/


; This makes string comparisons work as in JAWS 12 and older,
; making "RichEdit" match "RichEdit20W" for example.
;#pragma StringComparison partial

include "hjconst.jsh"
include "MSAAConst.jsh"
include "TeamTalk5Classic.jsm"
Include "common.jsm" ; message file


use "tt5_jcpdict.jsb"

Globals
String user,
String speaks,
String gRoom,
String sWinText,
String inform,
String talks,
String ourState,
String sVerb,
String ping,
String VoiceNone,
Int nMute,
Int nLeftDrive,
Int nRightDrive,
Int nVerb,
Int nConnect,
Int nFlagEvent,
Int nFlagVoice,
Int nFlagPush,
Int nUpdatedChannel,
Int nUser,
Int iWnd

const
vk = ";",
BM_CLICK = 0x00F5,
verb=";",
; Script revision number, spoken by JAWSKey+Q.
	Script_Revision = "96"

script scriptFileName()
scriptAndAppNames(formatString(msgScriptName, Script_Revision))
endScript

void function autoStartEvent()
toolbar__watchForChanges(10)
let speaks = iniReadString ("GraphicRes013232", "567387604", "", GetActiveConfiguration() + ".jgf")
Let speaks =StringSegment (speaks, "|", 1)
let gRoom = iniReadString ("GraphicRes013232", "4108880933", "", GetActiveConfiguration() + ".jgf")
Let gRoom =StringSegment (gRoom, "|", 1)
Let inform = ""
Let ourState = ""
SendMessage (GetAppMainWindow (GetCurrentWindow ()), 274, 61488, 0)
Let iWnd = 61488
; Flag
Let nFlagEvent = 0
Let nFlagVoice = 0
Let nFlagPush = 0
Let nUpdatedChannel = 0
autoStartEvent()
endFunction

script sayNextWord()
if !_moveByUnit("nextWord") then
	performScript sayNextWord()
endIf
endScript

script sayPriorWord()
if !_moveByUnit("priorWord") then
	performScript sayPriorWord()
endIf
endScript

int function _moveByUnit(string func)
if !isPCCursor() || util__isSpecialFocus() then
	return False
elif getObjectTypeCode() != WT_Edit then
	return False
endIf
var object rec let rec = chat4__record()
if !rec || rec("type") != "c" then
	; Private chat windows work fine already (4.3 Classic at least).
	return False
endIf

; RouteInvisibleToPC fails here by leaving the invisible cursor some 65,000
; pixels off the bottom right of the screen.  RouteJAWSToPC comes closer but
; can still end up at the wrong end of a word. The MSAA caret is available
; even in the read-only channel message window and is accurately located,
; so we use that. [DGL, 2012-03-17, JAWS 12.0.1170]

var int left, int right, int top, int bottom, int xc, int yc
; -8 is ObjID_Caret.
if !MSAA__windowRect(getFocus(), -8, left, right, top, bottom) then
	return False
endIf
let xc = (left+right) /2
let yc = (top+bottom) /2

saveCursor()  invisibleCursor()  saveCursor()
moveTo(xc, yc)
var int oldR let oldR = getRestriction()
setRestriction(RestrictWindow)
formatStringWithEmbeddedFunctions("<" +func +">")
setRestriction(oldR)
routePCToInvisible()
PCCursor()
let func = stringLower(func)
let func = stringReplaceSubstrings(func, "next", "")
let func = stringReplaceSubstrings(func, "prior", "")
pause()
formatStringWithEmbeddedFunctions("<say" +func +">")
return True
endFunction

int function ReadMsg(variant which, int isReviewSystem)
; Which should be +n (n forward), -n (n backward), n (message #n), "L" (last), or "C" (current).
var
	int msgno,
	int messageCount,
	int isJoinOrLeave,
	string buf
var object rec let rec = chat4__record()
let msgno = rec("reviewMessageNumber")

; First deal with room switches.
let messageCount = chat4__messageCount()
if messageCount < 0 then
	sayMessage(OT_Error, msgMessageError)
	return False
endIf
if !messageCount then
	let msgno = 0
	sayMessage(OT_User_Requested_Information, msgNoMessages)
	if isReviewSystem then
		rec.remove("reviewMessageNumber")
		rec.add("reviewMessageNumber", msgno)
	endIf
	return False
elif msgno < 1 then
	let msgno = 1
elif msgno > messageCount then
	let msgno = messageCount
endIf

if which == "L" then
	let msgno = messageCount
elif which == "C" then
	; Do nothing; leave msgno as is.
elif stringContains("+-", stringLeft(which, 1)) then
	let msgno = msgno +stringToInt(which)
else
	let msgno = stringToInt(which)
endIf

; Say the message after possible translation for welcome/goodbye style.
let buf = chat4__messageText(msgno)
applyChatTimePreference(buf)
let isJoinOrLeave = False

; the isReviewSystem part allows manually-requested messages to speak under all conditions.
var int chCycler let chCycler = 0
if isReviewSystem || !ChCycler || (ChCycler == 1 && isJoinOrLeave) || (ChCycler == 2 && !isJoinOrLeave) || (ChCycler == 3 && isJoinOrLeave) then
	if which == "c" then
		var string buf1 let buf1 = formatString(msgMessagePos, intToString(rec("reviewMessageNumber")), intToString(messageCount))
		let buf = formatString("%1: %2", buf1, buf)
	endIf
	sayMessage(OT_User_Requested_Information, buf)
elif !isReviewSystem && !isJoinOrLeave && (ChCycler == 3 || ChCycler == 4) then
	scheduleFunction("PlayMessageSound", 2)
endIf

if msgno < 1 then
	let msgno = 1
elif msgno > messageCount then
	let msgno = messageCount
endIf
if isReviewSystem then
	rec.remove("reviewMessageNumber")
	rec.add("reviewMessageNumber", msgno)
endIf
return True
endFunction

int function applyChatTimePreference(string byRef buf)
; Strip out the message time if requested by the user option for this.
if opt__get("chatMessageTimes") then
	return False
endIf
var int pos let pos = stringContains(buf, " ")
var string maybeTime let maybeTime = stringLeft(buf, pos-1)
if stringContains(maybeTime, ":")
&& stringContains("0123456789", stringLeft(maybeTime, 1))
&& stringContains("0123456789", stringRight(maybeTime, 1))
then
	let buf = stringChopLeft(buf, pos)
	return True
endIf
return False
endFunction

Script ReadFirstMessage()
ReadMsg("1", True)
EndScript
Script ReadLastMessage()
ReadMsg("L", True)
EndScript
Script ReadNextMessage()
ReadMsg("+1", True)
EndScript
Script ReadPriorMessage()
ReadMsg("-1", True)
EndScript
Script ReadCurrentMessage()
var object rec let rec = chat4__record()
var string buf let buf = chat4__messageText(rec("reviewMessageNumber"))
applyChatTimePreference(buf)
if isSameScript() then
	spellString(buf)
else
	ReadMsg("c", True)
endIf
EndScript
script readAbsoluteMessage(int idx)
if idx < 0 then
	let idx = chat4__messageCount() +idx +1
endIf
readMsg(idx, False)
endScript

script topOfFile()
if isPCCursor() && getFocus() && getFocus() == roomtree__currentTreeWindow() then
	roomtree__move("first", False)
	return
endIf
performScript topOfFile()
endScript

script selectFromTop()
if isPCCursor() && getFocus() && getFocus() == roomtree__currentTreeWindow() then
	roomtree__move("first", True)
	return
endIf
performScript selectFromTop()
endScript

script bottomOfFile()
if isPCCursor() && getFocus() && getFocus() == roomtree__currentTreeWindow() then
	roomtree__move("last", False)
	return
endIf
performScript bottomOfFile()
endScript

script selectToBottom()
if isPCCursor() && getFocus() && getFocus() == roomtree__currentTreeWindow() then
	roomtree__move("last", True)
	return
endIf
performScript selectToBottom()
endScript

script ControlUpArrow()
if isPCCursor() && getFocus() && getFocus() == roomtree__currentTreeWindow() then
	roomtree__move("prev", False)
	return
endIf
performScript ControlUpArrow()
endScript

script prevByAge()
if isPCCursor() && getFocus() && getFocus() == roomtree__currentTreeWindow() then
	roomtree__move("prev", True)
	return
endIf
sayCurrentScriptKeyLabel()
typeCurrentScriptKey()
endScript

script controlDownArrow()
if isPCCursor() && getFocus() && getFocus() == roomtree__currentTreeWindow() then
	roomtree__move("next", False)
	return
endIf
performScript controlDownArrow()
endScript

script nextByAge()
if isPCCursor() && getFocus() && getFocus() == roomtree__currentTreeWindow() then
	roomtree__move("next", True)
	return
endIf
sayCurrentScriptKeyLabel()
typeCurrentScriptKey()
endScript

script nextChat()
; Move to the next TeamTalk window containing a visible chat.
if !chat4__focusNext(1) then
	beep()
endIf
endScript

script prevChat()
; Move to the previous TeamTalk window containing a visible chat.
if !chat4__focusNext(-1) then
	beep()
endIf
endScript

script sayStatuses()
; Speak the statuses of toolbar icons.
; On one press, provide an abbreviated indication.
; On a double press, list all statuses completely.
toolbar__sayStatuses(isSameScript())
endScript

script DoEnter()
; Makes Enter join a room if focus is on a room line or on someone within one.
var handle hwnd let hwnd = getFocus()
if GetWindowClass (GetCurrentWindow ()) == "RichEdit20W then"
PlaySound (GetSharedSettingsDirectory ()+"\\SOUNDS\\ascend.wav")
endif
if getWindowClass(hwnd) == "SysTreeview32" then
	if _DoubleClick(hwnd) then
		return
	endIf
	beep()
endIf
typeCurrentScriptKey()
sayCurrentScriptKeyLabel()
endScript

int function _DoubleClick(handle hwnd)
; Double-click the focused tree item.
; For a room node, this joins the room if "Double click enters a room" is set in TeamTalk.
; For a member, this opens a chat window.
; Returns True on success and False on error.
var object o let o = MSAA__objectFromWindow(hwnd)
if !o then
	return False
endIf
var int childID let childID = o.accFocus +0
if !childID then
	; Even if this means focus is just on the tree itself,
	; that's not a room to join or a member to talk to.
	return False
endIf
; Even though we're using winclick to avoid having to mess with the mouse pointer,
; we probably have to have the node actually on screen.
; TODO: Could improve this to allow for partial visibility.
var int accState let accState = o.accState(childID)
if (accState & (State_System_Invisible | State_System_OffScreen)) then
	return False
endIf
; Click the center of the node.
var int left, int right, int top, int bottom
if !MSAA__rect(o, childID, left, right, top, bottom) then
	return False
endIf
var int x let x = (left +right) /2
var int y let y = (top +bottom) /2
winclick__post(hwnd, "LeftDblClick", x, y)
; Send seems to return 0 even on success.
return True
endFunction

void function jcp__setCustomProps(variant byRef props, string sOrigin)
; Set any custom properties required per situation.
; props is empty on entry, can be modified, and must be taken by reference.
; Properties understood by default (property name case insignificant):
; Properties used by both Braille and speech:
;	Name, type, state
;	containerName, containerType
;	value, position, dlgText
;	itemPos, itemCount (convenient way to compose position)
;	Hotkey (used in SayTutorialHelpHotKey)
;	subtypeCode (default type for speech and SayTutorialHelp and returned by BrailleCallbackObjectIdentify for this control)
;	attributes (CTRL_* constants from hjconst.jsh defining the control's states)
;	beforeField, afterField (text to speak/Braille before/after this field)
; Speech-only properties:
;	tutor, beforeTutor (tutor overrides the normal tutor message for the subtype)
; Braille-only properties:
;	dlgPageName, contextHelp, time, level
; Special properties internal to this system:
;	Mask:  Masks all special handling (use with caller name, e.g., set(props, "Mask.SayLine", True)).
;	Origin:  The origin to use to get default property values with getObject* functions (usually the sOrigin value used to create this property set)
; Note that SubtypeCode will set Type if Type is not set already; same for Attributes and State.
; sOrigin values that must be handled:
;	focus, current
;	subtype<subtype> (for BrailleAddObject* functions)
;	hwnd<handle>
; Other possible sOrigin values:
;	point<x>,<y>:  Like current but at a specific point
;	obj[<level>]  (level 0 is current)
;	hwnd<handle>/{nav<id>|sdm<id>|MSAA<id>}:  Nav* function ID, SDM ID, or MSAA ID within a window
;	point<x>,<y>:{hwnd|nav|SDM|MSAA}:  Like the extended hwnd syntax but from a specific point
var handle hwnd, object o, int childID
if isVirtualPCCursor() then
	return
endIf
jcp__setByOrigin(sOrigin, "hm", hwnd, o, childID)
if util__isSpecialFocus() then
	return
endIf
if ttutil__isMainWindow(getTopLevelWindow(hwnd)) then
	_labelMain(props, hwnd)
endIf
endFunction

int function _labelMain(variant byRef props, handle hwnd)
; Label the sliders and other items in the main window.
var string wclass let wclass = getWindowClass(hwnd)
var string name
if wclass == "Edit" then
	jcp__set(props, "Name", msgChannelMessageInput)
	return True
elif wclass == "RichEdit20W" then
	jcp__set(props, "Name", msgChannelMessageHistory)
	return True
elif wclass == "SysTreeView32" then
	var string flags let flags = roomtree__iconStatus()
	if flags then
		let name = getObjectValue()
		; That includes flags because of a getObjectValue() override in this file.
		jcp__set(props, "Value", name)
	endIf
	return True
endIf
if getWindowTypeCode(hwnd) != WT_Slider then
	return False
endIf
var handle h1 let h1 = getFirstWindow(hwnd)
var int i let i = 0
var string names let names = msgSliderNames
while h1
	if getWindowTypeCode(h1) == WT_Slider then
		let i = i +1
	endIf
	if h1 == hwnd then
		let name = stringSegment(names, "|", i)
		jcp__set(props, "Name", name)
		return True
	endIf
	let h1 = getNextWindow(h1)
endWhile
return False
endFunction

string function getObjectValue(/*?*/ int alwaysUseMSAA, int nLevel)
var string val let val = getObjectValue(alwaysUseMSAA, nLevel)
if nLevel != 0 then
	return val
endIf
if !util__isSpecialFocus() && isPCCursor()
&& getWindowClass(getFocus()) == "SysTreeView32" then
	var string flags let flags = roomtree__iconStatus()
	var string buf let buf = ""
	if stringContains(flags, "s") then
		let buf = buf +" " +msgSending
	endIf
	if stringContains(flags, "i") then
		let buf = buf +" " +msgIdle
	endIf
	if stringContains(flags, "o") then
		let buf = buf +" " +msgOnline
	endIf
	let val = stringTrimTrailingBlanks(val)
	let buf = stringChopLeft(buf, 1)
	let val = string__join(val, " - ", buf)
endIf
return val
endFunction


;================================= Module chat4 ================================
; ChatOutput window interface for TeamTalk4Classic
; This is an iTextDocument interface.
;
; Interface:
;	chat4__record(): Get the data record for the current or a specific chat.
;	chat4__messageIndex() and chat4__messageCount(): Position in and count of messages.
;	chat4__messageText([index]): Text of current message or message with given 1-based index.
;	chat4__setPos(index):  Set the "current" message.
;	chat4__move(delta):  Move delta (plus or minus) messages from current position.
; Attempting to move outside the valid message range causes a beep and stays at the border encountered..
;
; Author:  Doug Lee


globals
; Records for chat windows, by top-level window handle (Scripting.Dictionary object).
; Fields:
;	hTop: The top-level window (same as the key for this record).
;	hwnd, oDoc: The actual chat window and its RichEdit document object.
;	type: The type of the window: c=channel, p=private.
;	unit: The TOM_* text unit for moving among messages.
;	firstStart, firstIdx: The starting position and index of the first "real" message.
;	curStart: The starting position of the range for the current text unit.
;	range: A Range object set by chat4___messageIndex() and reused by chat4__messageText() to get message text.
;	rows: Current number of rows in the message text (see chat4___rows()).
; Note: The caller of this module manages a reviewMessageNumber field as well.
; TODO: That should be handled in here.
	object chat4___chats

object function chat4__record(/*?*/ handle hwnd)
; Get or make the record for the current or given chat.
; Allows fields to be filled in that were missed at first.
if !hwnd || hwnd == 0 then
	let hwnd = getFocus()
endIf
var handle hTop let hTop = getTopLevelWindow(hwnd)
if !hTop || !isWindowVisible(hTop) then
	return
endIf
var object rec let rec = chat4__winrecs__record(chat4___chats, hTop)
if !rec then
	return
endIf

; Build or finish the record.

; Top-level window and navigation unit.
if !rec.exists("hTop") then
	rec.add("hTop", hTop)
	; Messages appear to be considered paragraphs.
	rec.add("unit", 4)  ; TOMParagraph
endIf

; Chat window and chat type.
if !rec.exists("hwnd") then
	; First try for a channel window.
	let hwnd = findWindow(hTop, "RichEdit20W", "")
	if hwnd then
		rec.add("type", "c")
	else
		; Then for a private chat window.
		let hwnd = findWindow(hTop, "RichEdit20A", "")
		if hwnd then
			rec.add("type", "p")
		endIf
	endIf
	if !hwnd then
		chat4__winrecs__remove(chat4___chats, hTop)
		return
	endIf
	rec.add("hwnd", hwnd)
endIf

; Chat window's RichEdit interface.
var object o
if !rec.exists("oDoc") then
	let o = getRichEditDocument(hwnd)
	if !o then
		return
	endIf
	rec.add("oDoc", o)
else
	let o = rec("oDoc")
endIf

; Row count for chat.
if !rec.exists("rows") then
	rec.add("rows", chat4___rows(hwnd))
endIf

if !rec.exists("firstStart") then
	rec.add("firststart", 0)
	rec.add("curstart", 0)
endIf

; Set rec("firststart") if necessary and possible.
if !rec("firststart") then
	if chat4___size(rec("hwnd")) < 10 then
		return
	endIf
	var object oRange let oRange = o.Range(0, 0)
	var int safety let safety = 0
	oRange.Expand(rec("unit"))
	while oRange.End -oRange.Start < 4 && safety < 10
		oRange.Collapse(0)
		oRange.Move(rec("unit"), 1)
		oRange.Expand(rec("unit"))
		let safety = safety +1
	endWhile
	if safety >= 10 then
		rec.remove("firststart")   rec.add("firststart", 0)
		rec.remove("firstidx")   rec.add("firstidx", 0)
		;sayInteger(chat4___size(rec("hwnd")))
	else
		rec.remove("firststart")   rec.add("firststart", oRange.Start)
		rec.remove("firstidx")   rec.add("firstidx", oRange.getIndex(rec("unit")))
	endIf
endIf
return rec
endFunction

int function chat4__clean()
; Clean out invalid chat windows.
; Returns the number of windows removed
var int nremoved let nremoved = chat4__winrecs__clean(chat4___chats)
var string hwnds let hwnds = chat4__winrecs__list(chat4___chats)
if !hwnds then
	return 0
endIf
var int n let n = stringSegmentCount(hwnds, "|")
var int i let i = 1
var handle hwnd
while i <= n
	let hwnd = stringToHandle(stringSegment(hwnds, "|", i))
	if isWindowDisabled(hwnd) then
		chat4__winrecs__remove(chat4___chats, hwnd)
		let nremoved = nremoved +1
	endIf
	let i = i +1
endWhile
return nremoved
endFunction

int function chat4___rows(handle hwnd)
; Return the number of rows of material in the given message window.
; This is used to detect changes.
var int lcount let lcount = sendMessage(hwnd, 0xBA, 0, 0)  ; EM_GetLineCount
return lcount
endFunction

int function chat4___size(handle hwnd)
; Return the "size" of the material in the given message area window.
return sendMessage(hwnd, 0x0E, 0, 0)  ; WM_GetTextLength
endFunction

int function chat4___messageIndex(int start, /*?*/ variant which)
; Return the 1-based index of the message whose start position is given, or 0 on failure..
; A start position of -1 means the last message.
; Blank lines at the top and bottom of the buffer are avoided.
; Which can bbe a window handle, a record object, or omitted.
; This function is also used internally to set the rec("range") object.
var object rec
if !which then
	; Record for currently focused chat.
	let rec = chat4__record()
elif intToString(which) then
	; Record for given window handle.
	let rec = chat4__record(which)
else
	; Record was passed directly.
	let rec = which
endIf
if !rec then
	return 0
endIf

var object oRange
if start == -1 then
	; Find the end of the buffer.
	let oRange = rec("oDoc").Range(0, 0)
	oRange.Expand(6)  ; 6 is TOMStory
	; A blank line tends to remain at the end of the buffer.
	; -2 here avoids it.
	let start = oRange.End -2
endIf
; This avoids the blank line before the first message.
if start < rec("firststart") then
	let start = rec("firststart")
endIf
let oRange = rec("oDoc").Range(start, start)
oRange.Expand(rec("unit"))
rec.remove("range")   rec.add("range", oRange)
return oRange.getIndex(rec("unit")) -rec("firstidx") +1
endFunction

int function chat4__messageIndex(/*?*/ handle hwnd)
; Returns the 1-based index of the last-read message.
var object rec let rec = chat4__record(hwnd)
return chat4___messageIndex(rec("curstart"), rec)
endFunction

int function chat4__messageCount(/*?*/ handle hwnd)
; Returns the number of messages in the buffer.
; A return of -1 indicates an error.
; TODO: Error checking is not here yet.
var object rec let rec = chat4__record(hwnd)
return chat4___messageIndex(-1, rec)
endFunction

string function chat4__messageText(/*?*/ int idx, handle hwnd)
; Return the text of the current or indicated message.
; Negative numbers index from the end.
; The current message position is not changed.
var object rec let rec = chat4__record(hwnd)
var string txt
if idx then
	if idx < 0 then
		chat4___messageIndex(-1, rec)
		; sets rec("range") to the last message.
		if idx != -1 then
			chat4___move(idx+1, rec)
		endIf
	else  ; positive idx.
		chat4___messageIndex(rec("firststart"), rec)
		; sets rec("range") to the first message.
		if idx != 1 then
			chat4___move(idx-1, rec)
		endIf
	endIf
else
	; Current message wanted.
	chat4___messageIndex(rec("curstart"), rec)
	; rec("range") set and we don't have to move anywhere from here.
endIf
let txt = rec("range").Text
return stringTrimLeadingBlanks(stringTrimTrailingBlanks(txt))
endFunction

int function chat4___move(int delta, /*?*/ object rec)
; Move delta messages from the current position of rec("range").
; rec("range") must be set first.
; Note that this is from rec("range"), not rec("curstart").
; Returns the distance moved.
if !delta then
	return 0
endIf
if !rec then
	let rec = chat4__record()
	if !rec then
		return 0
	endIf
endIf
var int curidx let curidx = chat4___messageIndex(rec("range").Start, rec)
rec("range").Collapse(-1)  ; -1 is TOMTrue, for collapse to start.
if chat4___trimDelta(delta, curidx, rec) then
	beep()
endIf
rec("range").Move(rec("unit"), delta)
rec("range").Expand(rec("unit"))
var int result let result =
util__abs(curidx -chat4___messageIndex(rec("range").Start))
return result
endFunction

int function chat4___trimDelta(int byRef delta, int curidx, object rec)
; Forces delta not to move outside the valid messages.
; Returns True if delta is changed and False if not.
if delta < 0 then
	if curidx+delta < 1 then
		let delta = 1-curidx
		return True
	endIf
else
	var object oRange let oRange = rec("range").duplicate()
	var int lastidx let lastidx = chat4___messageIndex(-1, rec)
	rec.remove("range")   rec.add("range", oRange)
	if curidx +delta > lastidx then
		let delta = lastidx -curidx
		return True
	endIf
endIf
return False
endFunction

int function chat4__move(int delta, /*?*/ handle hwnd)
; Move delta messages from the current position.
; Returns the number (absolute value) of messages moved.
var int result
var object rec let rec = chat4__record(hwnd)
chat4___messageIndex(rec("curstart"), rec)
; That sets rec("range") to reflect the current position.
let result = chat4___move(delta, rec)
rec.remove("curstart")   rec.add("curstart", rec("range").Start)
return result
endFunction

int function chat4__setPos(int idx, /*?*/ handle hwnd)
; Set the current message to the given index.
; Negative indices count from the end.
var object rec let rec = chat4__record(hwnd)
chat4__messageText(idx, rec("hTop"))
rec.remove("curstart")   rec.add("curstart", rec("range").Start)
endFunction

int function chat4__focusNext(int delta)
; Focus the next or previous chat.
; Delta should be 1 for next and -1 for previous.
chat4___registerAllChatWindows()
var string hwnds let hwnds = chat4__winrecs__list(chat4___chats, "n")
if !hwnds then
	return False
endIf
var handle hTop let hTop = getTopLevelWindow(getFocus())
if !hTop then
	return False
endIf
var int idx let idx = stringSegmentIndex(hwnds, "|", intToString(hTop), True)
var int n let n = stringSegmentCount(hwnds, "|")
if !idx then
	if delta > 0 then
		let idx = n
	else
		let idx = 1
	endIf
endIf
let idx = idx +delta
if idx < 1 then
	let idx = idx +n
elif idx > n then
	let idx = idx -n
endIf
var handle hwnd let hwnd = stringToHandle(stringSegment(hwnds, "|", idx))
if hwnd == hTop then
	sayMessage(OT_Error, chat4___msgNoMoreWindows)
	return True
endIf
setFocus(hwnd)
var int safety let safety = 50
while safety
	if getTopLevelWindow(getFocus()) == getTopLevelWindow(hwnd) then
		let safety = 1
	elif !getFocus() && getForegroundWindow() == getTopLevelWindow(hwnd)
	&& MSAA__windowState(getForegroundWindow(), 0, ObjID_Window) & State_System_Invisible then
		typeKey("Enter")
		let safety = 1
	else
		pause()
	endIf
	let safety = safety -1
endWhile
return True
endFunction

void function chat4___registerAllChatWindows()
; Build or update the list of all valid chat windows.
var string thisOwner let thisOwner = getWindowOwner(getTopLevelWindow(getFocus()))
var handle hwnd let hwnd = findTopLevelWindow("#32770", "")
while hwnd
	if isWindowVisible(hwnd) && !isWindowDisabled(hwnd)
	&& stringCompare(getWindowOwner(hwnd), thisOwner, True) == 0 then
		chat4__record(hwnd)
	endIf
	let hwnd = getNextWindow(hwnd)
endWhile
chat4__clean()
endFunction


;----------------------- Module winrecs (from winrecs10) -----------------------
; A manager for data structures (records) indexed by a window handle.
;
; Author:  Doug Lee of SSB BART Group

const
; The hash key used to store the list of window handles,
; because JAWS can't use the keys() method of Scripting.Dictionary objects.
	chat4__winrecs___HwndListKey = "__hwnds__"

object function chat4__winrecs__record(object byRef recs, handle hwnd)
; Return the record for the given window.
; Registers this window in the set if necessary.
; Also removes stale records for windows that no longer exist.
; The caller must handle record structure creation.
; Suggestion: Use an "hwnd" property so there's a uniform way to test for a new record.
chat4__winrecs__clean(recs)
if !hwnd then
	return
endIf
var string shwnd let shwnd = intToString(hwnd)
var object rec
if !recs then
	let recs = util__newCollection()
endIf
if !recs.exists(shwnd) then
	; We have to make a new record for this one.
	var string hwnds let hwnds = recs(chat4__winrecs___HwndListKey)
	if !hwnds then
		let hwnds = shwnd
	else
		let hwnds = formatString("%1|%2", hwnds, shwnd)
	endIf
	recs.remove(chat4__winrecs___HwndListKey)
	recs.add(chat4__winrecs___HwndListKey, hwnds)
	let rec = util__newCollection()
	recs.add(shwnd, rec)
	; Caller must detect that this record is empty and build it.
else
	let rec = recs(shwnd)
endIf
return rec
endFunction

int function chat4__winrecs__remove(object recs, handle hwnd)
; Remove a window's record.
var string shwnd let shwnd = intToString(hwnd)
if !recs.exists(shwnd) then
	return False
endIf
recs.remove(shwnd)
var string hwnds let hwnds = recs(chat4__winrecs___HwndListKey)
if !string__segmentRemove(hwnds, "|", shwnd) then
	chat4__winrecs__log__write("Failed to remove handle " +shwnd +" from the window list.")
	return True
endIf
recs.remove(chat4__winrecs___HwndListKey)
recs.add(chat4__winrecs___HwndListKey, hwnds)
return True
endFunction

int function chat4__winrecs__clean(object recs)
; Remove from recs any records for windows that no longer exist.
; Returns the number of records removed.
var string hwnds let hwnds = recs(chat4__winrecs___HwndListKey)
var int n let n = stringSegmentCount(hwnds, "|")
var int nremoved let nremoved = 0
var string shwnd, handle hwnd
while n
	let shwnd = stringSegment(hwnds, "|", n)
	let hwnd = stringToHandle(shwnd)
	if !getWindowStyleBits(hwnd) then
		if chat4__winrecs__remove(recs, hwnd) then
			let nremoved = nremoved +1
		endIf
	endIf
	let n = n -1
endWhile
return nremoved
endFunction

string function chat4__winrecs__list(object recs, /*?*/ string how)
; Return a |-delimited list of the registered window handles.
; By default, handles are returned in the order in which they were added to the set.
; If how is "n", they are returned in numeric order.
; If how is anything else, it is considered a function name or call.
; See util__compare() for details.
var string hwnds let hwnds = recs(chat4__winrecs___HwndListKey)
if !how then
	return hwnds
endIf
; Kludge: Make XList out of hwnds.
let hwnds = "|" +hwnds +"|"
let hwnds = chat4__winrecs__xlist__sort(hwnds, how, False)
; Remove the extra delimiters added above.
let hwnds = stringChopLeft(stringChopRight(hwnds, 1), 1)
return hwnds
endFunction


;__________________________________ Module log _________________________________
; Allow logging to a text file.
;
; Interface:  chat4__winrecs__log__open, chat4__winrecs__log__isOpen, chat4__winrecs__log__close, chat4__winrecs__log__write, chat4__winrecs__log__setUserID.
; Basic usage: chat4__winrecs__log__open() from AutoStartEvent, chat4__winrecs__log__write("...") to log something.
; Usage:
;	- Call chat4__winrecs__log__open() from AutoStartEvent.
;	- Logging will occur by default only if the log file already exists (good for debugging during development).
;	- Call as chat4__winrecs__log__open("", True) if you want logging to occur even if the log file doesn't already exist.
;	- Pass a file name instead of "" for a custom file name (rarely needed).
;	- Call chat4__winrecs__log__write("Oops I saw an error") to log something.
;	- Use chat4__winrecs__log__write("Oops this error needs to speak too", True) to speak what is being logged.
;	- Call chat4__winrecs__log__close() to disable logging (rarely needed).
;
; Author:  Doug Lee of SSB BART Group

; Constants for chat4__winrecs__log__write and related functions.
; Not localized because log text is not localized.

; Used as a user ID when none is found.
const chat4__winrecs__log___msgNoName = "<no name>"
; What "name" is used for the caller string when JAWS is too old for getScriptCallStack().
; %1 is the JAWS version.
const chat4__winrecs__log___msgJAWSTooOld = "(caller unknown; JAWS version (%1) too old)"
; Name used for the caller when it can't be determined for an unknown reason,
; probably a getScriptCallStack return value format surprise.
const chat4__winrecs__log___msgNoCaller = "(caller unknown)"

globals
; FileSystemObject and log stream object obtained from it.
	object chat4__winrecs__log___fso,
	object chat4__winrecs__log___logstream,
; Name of file.
	string chat4__winrecs__log___fileName,
	int chat4__winrecs__log___canCreate,
	string chat4__winrecs__log___userID,
	string chat4__winrecs__log___machineID

int function chat4__winrecs__log___setup()
; Internal function called to set things up when necessary before any log text is generated.
if !stringIsBlank(chat4__winrecs__log___userID) then
	return True
endIf
let chat4__winrecs__log___fso = createObjectEx("Scripting.FileSystemObject", False)
if !chat4__winrecs__log___fso then
	return False
endIf
let chat4__winrecs__log___machineID = ""
var object oNet let oNet = createObjectEx("WScript.Network", False)
if oNet then
	let chat4__winrecs__log___machineID = oNet.ComputerName
endIf
if stringIsBlank(chat4__winrecs__log___machineID) then
	let chat4__winrecs__log___machineID = chat4__winrecs__log___msgNoName
endIf
let chat4__winrecs__log___userID = getJAWSUserName()
if stringIsBlank(chat4__winrecs__log___userID) && oNet then
	let chat4__winrecs__log___userID = oNet.UserName
endIf
if stringIsBlank(chat4__winrecs__log___userID) then
	let chat4__winrecs__log___userID = chat4__winrecs__log___msgNoName
endIf
return True
endFunction

int function chat4__winrecs__log__setUserID(string id)
; Optional means of setting a custom user ID.
; If not called, the value of getJAWSUserName() is used.
let chat4__winrecs__log___userID = id
return True
endFunction

int function chat4__winrecs__log__close()
; Close the log file.
chat4__winrecs__log___close()
let chat4__winrecs__log___fileName = ""
return True
endFunction

int function chat4__winrecs__log___setFileName(string name)
; Set the log file name.
; The name can be passed; otherwise <basename>.log is used.
; <basename> is getActiveConfiguration() or, failing that, the basename of getAppFileName().
; The log file is located in the JAWS user folder.
; Helper for chat4__winrecs__log__open().
var string basename
let basename = name
if stringRight(basename,4) == ".log" then
	let basename = stringChopRight(basename, 4)
endIf
if !basename then
	let basename = getActiveConfiguration()
endIf
if !basename then
	let basename = getAppFileName()
	var int pos let pos = stringContainsFromRight(basename, ".")
	if pos > 0 then
		let basename = stringLeft(basename, pos-1)
	endIf
endIf
let chat4__winrecs__log___fileName = getJAWSSettingsDirectory() +"\\" +basename
if stringRight(chat4__winrecs__log___fileName, 4) != ".log" then
	let chat4__winrecs__log___fileName = chat4__winrecs__log___fileName +".log"
endIf
return True
endFunction

int function chat4__winrecs__log__open(/*?*/ string name, int canCreate)
; Open and optionally create the log file.
; The name can be passed; otherwise <basename>.log is used.
; <basename> is getActiveConfiguration() or, failing that, the basename of getAppFileName().
; The log file is located in the JAWS user folder.
; Call from autoStartEvent, or logging will not occur.
; Returns True if the file is opened and False if not.
if !chat4__winrecs__log___setup() then
	return False
endIf
let chat4__winrecs__log___canCreate = canCreate
chat4__winrecs__log__close()
if !chat4__winrecs__log___setFileName(name) || !chat4__winrecs__log___open() then
	let chat4__winrecs__log___fileName = ""
	return False
endIf
return True
endFunction

int function chat4__winrecs__log___open()
; Raw open/create.  Sets chat4__winrecs__log___logstream.
; Returns True on success and False on failure.
if chat4__winrecs__log___logstream then
	return True
endIf
if fileExists(chat4__winrecs__log___fileName) then
	let chat4__winrecs__log___logstream = chat4__winrecs__log___fso.openTextFile(chat4__winrecs__log___fileName, 8)  ; Append
elif chat4__winrecs__log___canCreate then
	let chat4__winrecs__log___logstream = chat4__winrecs__log___fso.CreateTextFile(chat4__winrecs__log___fileName)
endIf
if !chat4__winrecs__log___logstream then
	return False
endIf
scheduleFunction("chat4__winrecs__log___close", 30)
return True
endFunction

int function chat4__winrecs__log___close()
; Raw cloase; unsets chat4__winrecs__log___logstream.
var variant null
if !chat4__winrecs__log___logstream then
	return True
endIf
chat4__winrecs__log___logstream.Close()
let chat4__winrecs__log___logstream = null
return True
endFunction

int function chat4__winrecs__log___isOpen()
; Returns True if the log file is actually open and False if not.
if chat4__winrecs__log___logstream then
	return True
endIf
return False
endFunction

int function chat4__winrecs__log__isOpen()
; Returns True if the log file is open and False if not.
; "Open" here means can be written to but may not necessarily mean the file is physically opened at the moment.
if chat4__winrecs__log___fileName then
	return True
endIf
return False
endFunction

string function chat4__winrecs__log___getStamp()
; Returns a formatted date/time stamp for the log file.
var string tm let tm = sysGetTime("HH:mm:ss")
var string dt let dt = sysGetDate("ddd MMM dd yyyy")
return formatString("%1 %2", tm, dt)
endFunction

int function chat4__winrecs__log__write(string s, /*?*/ int alsoSpeak)
; Write a string to the log file, and also optionally speak it.
if alsoSpeak then
	sayMessage(OT_Error, s)
endIf
if !chat4__winrecs__log__isOpen() then
	return False
endIf
if !chat4__winrecs__log___open() then
	return False
endIf
var string buf let buf = formatString(
	"*** %1 %2@%3 (%4)\n%5:\n",
	chat4__winrecs__log___getStamp(), chat4__winrecs__log___userID, chat4__winrecs__log___machineID, getAppFileName(),
	chat4__winrecs__log___getCaller()
) +s +"\n\n"
; Standardize any CR/LF characters in buf before writing.
let buf = stringReplaceChars(buf, "\r\n", "\n")
let buf = stringReplaceSubstrings(buf, "\n", "\r\n")
chat4__winrecs__log___logstream.write(buf)
return True
endFunction

string function chat4__winrecs__log___getCaller()
; Get and return the caller of this logging system.
; Helper for chat4__winrecs__log__write().
if getJFWVersion() < 1000000 then
	var string jver let jver = stringSegment(util__getJAWSInfo(), "|", 1)
	return formatString(chat4__winrecs__log___msgJAWSTooOld, jver)
endIf
var string stack let stack = util__callStackInfo(2, "s", 2)
if !stringIsBlank(stack) then
	return stack
endIf
return formatString("%1\nStack begin:\n%2\nStack end.",
	chat4__winrecs__log___msgNoCaller,
	util__vcast(getScriptCallStack())
)
endFunction

string function chat4__winrecs__log___splitOff(string byRef s, string sep, /*?*/ int fromRight)
; Split what's before sep off and return it.
; If fromRight is True, split everything before the last sep off and return that.
; Returns by reference s with both that and sep removed.
var string s1
var int pos
if fromRight then
	let pos = stringContainsFromRight(s, sep)
else
	let pos = stringContains(s, sep)
endIf
if pos then
	let s1 = stringLeft(s, pos-1)
	let s = stringChopLeft(s, pos+stringLength(sep)-1)
else
	let s1 = s
	let s = ""
endIf
return s1
endFunction

;________________________________ End module log _______________________________


;_________________________________ Module xlist ________________________________
; XList, the arbitrary list (array and hash emulation) module.
;
; Terminology:
;	XList - An arbitrary list of items (an array, basically).
;	XLHash - An XList in which odd-numbered items are keys and even-numbered items are their values.
;
; Implementation Notes:
; XLists always begin and end with a delimiter; a null list is just a single delimiter character.
; This means it is NOT possible to use boolean (if sl) or length (if stringLength(sl)) to check for emptiness; use isEmpty or itemCount.
; Objects can not be stored in an XList directly, and attempting this will cause an error message to speak.
; Translation Notes:  The only translatable strings appear in SayMessage calls.
;
; Author:  Doug Lee
;
; Revision History:  See version control logs.

const
; Indexes used to request the character tables supported by this module.
	chat4__winrecs__xlist___CharANSI = 1,
; The corresponding character tables.
; The double quote (") and backslash characters must be escaped.
	chat4__winrecs__xlist___CharANSICharList = "\1\2\3\4\5\6\7\8\9\10\11\12\13\14\15\16\17\18\19\20\21\22\23\24\25\26\27\28\29\30\31 !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~\127\128\129\130\131\132\133\134\135\136\137\138\139\140\141\142\143\144\145\146\147\148\149\150\151\152\153\154\155\156\157\158\159\160\161\162\163\164\165\166\167\168\169\170\171\172\173\174\175\176\177\178\179\180\181\182\183\184\185\186\187\188\189\190\191\192\193\194\195\196\197\198\199\200\201\202\203\204\205\206\207\208\209\210\211\212\213\214\215\216\217\218\219\220\221\222\223\224\225\226\227\228\229\230\231\232\233\234\235\236\237\238\239\240\241\242\243\244\245\246\247\248\249\250\251\252\253\254\255"

string function chat4__winrecs__xlist__sort(string xl, string how, int sortDescending)
; Sorts xl and returns the result.
; How is as for util__compare().
; This is a stable sort; equal items are not swapped.
; TODO:  The algorithm is slow:  sorting 200 items took nine seconds on a Sony Vaio 1.2 GHz Centrino machine (Windows XP, JAWS 7.10, DGL, 2006-08-25).
var
	string xl1,
	string tmp, string firstVal, int firstIdx,
	int i, int n, int cmp
; Using the below trick instead of chat4__winrecs__xlist__new() preserves the separator from xl on return.
let xl1 = stringLeft(xl, 1)
let n = chat4__winrecs__xlist__itemCount(xl)
if !n then
	return xl1
elif n == 1 then
	; The commented-out approach guarantees an XList structure return;
	; the faster plan, in effect, allows single-item strings to pass through unchecked.
	;chat4__winrecs__xlist__addItem(xl1, chat4__winrecs__xlist__getItem(xl, 1))
	;return xl1
	return xl
endIf

; We have two or more items.
while n > 1
	let firstVal = chat4__winrecs__xlist__getItem(xl, 1)
	let firstIdx = 1
	let i = 2
	while i <= n
		let tmp = chat4__winrecs__xlist__getItem(xl, i)
		let cmp = util__compare(firstVal, tmp, how)
		if sortDescending then
			let cmp = cmp * (-1)
		endIf
		if cmp > 0 then
			let firstVal = string__ref(tmp)
			let firstIdx = i
		endIf
		let i = i +1
	endWhile
	; firstVal and firstIdx are the value and index of the "first" value, respectively.
	; "First" means smallest for ascending and greatest for descending sorts.
	chat4__winrecs__xlist__addItem(xl1, firstVal)
	chat4__winrecs__xlist__delete(xl, firstIdx, 1)
	let n = n -1
endWhile
; What's left is the max item.
chat4__winrecs__xlist__addItem(xl1, chat4__winrecs__xlist__getItem(xl, 1))
return xl1
endFunction

int function chat4__winrecs__xlist__delete(string byRef lst, int i, int n)
; Deletes n elements starting at element i.
var
	int pos1, int pos2
let pos1 = chat4__winrecs__xlist___indexPos(lst, i)
if !pos1 || pos1 == stringLength(lst)+1 then
	return False
endIf
let pos2 = chat4__winrecs__xlist___indexPos(lst, i+n)
if !pos2 || pos2 == stringLength(lst)+1 then
	let lst = stringLeft(lst, pos1-1)
	return True
endIf
let lst = string__cat(stringLeft(lst, pos1-1), stringChopLeft(lst, pos2-1))
return True
endFunction

int function chat4__winrecs__xlist__addItem(string byRef lst, variant item)
; Append item to lst.
; If lst is null, it will be created first; but if it is non-null and not an XList, a mess is likely.
; Returns True on success and False on failure.
; On failure, lst is not changed (except for maybe being initialized if null).
if !stringLength(item) && item then
	sayMessage(OT_No_Disable, "Can't store an object in an XList")
	return False
endIf
if !chat4__winrecs__xlist___settleDelim(lst, item, 1) then
	return False
endIf
let lst = string__cat(lst, string__cat(item, stringLeft(lst, 1)))
return True
endFunction

variant function chat4__winrecs__xlist__getItem(string lst, int i)
; Returns the ith element of lst.
; If i is greater than the item count, the null string is returned.
var string val
if i < 1 || !stringLength(lst) then
	return ""
endIf
let val = stringSegment(lst, stringLeft(lst, 1), i+1)
return val
endFunction

int function chat4__winrecs__xlist__itemCount(string lst)
; Returns the number of items in lst.
; If lst is not a proper XList, returns 0.
if !stringLength(lst) then
	return 0
endIf
return string__segmentCount(lst, stringLeft(lst, 1)) -2
endFunction

int function chat4__winrecs__xlist___settleDelim(string byRef lst, string item, int dlen)
; Internal function to manage the changing of delimiters as needed.
; Must be called whenever material is being added to an XList.
; dlen is the length of lst's current delimiter (1 except when sent via fromJAWSList).
; Returns True if a valid delimiter is in place on exit and False if there is no safe delimiter available.
; A "valid" delimiter is a one-character delimiter not found in lst (except as the current delimiter) or item.
var
	string delim,
	int c0, int c, string ch,
	int done
if !stringLength(lst) then
	let lst = chat4__winrecs__xlist__new()
endIf
let delim = stringLeft(lst, dlen)
if stringContains(item, delim) || dlen != 1 then
	let c = chat4__winrecs__xlist__charToANSI(stringLeft(lst, 1))
	let c0 = c
	let done = False
	while !done
		let c = c +1
		if c > 254 then
			; 255 may not write correctly to a file, and 0 is not possible in JAWS.
			let c = 1
		elif c >= 8 && c <= 13 then
			; backspace, tab, LF, RF(?), FF, CR:
			; These characters are unfriendly when written to a text file, so they are avoided.
			let c = 14
		elif c == 26 || c == 27 then
			; DOS/Windows EOF, ESC
			let c = 28
		endIf
		let ch = chat4__winrecs__xlist__charFromANSI(c)
		if c == c0 then
			; We've come all the way around; there is no safe delimiter for this situation.
			beep()
			sayMessage(OT_No_Disable, "xlist:  Out of delimiters")
			return False
		elif !stringContains(lst, ch) && !stringContains(item, ch) then
			; We have a winner!
			chat4__winrecs__xlist___setDelim(lst, ch, dlen)
			let delim = ch
			let done = True
		endIf
	endWhile
endIf
return True
endFunction

int function chat4__winrecs__xlist__charToANSI(string c)
; Returns the ANSI code for the first character in c, or 0 if c is null.
return chat4__winrecs__xlist___getCharCode(c, chat4__winrecs__xlist___CharANSI)
endFunction

string function chat4__winrecs__xlist__charFromANSI(int i)
; Returns the character corresponding to the given ANSI code.
return chat4__winrecs__xlist___getChar(i, chat4__winrecs__xlist___CharANSI)
endFunction

string function chat4__winrecs__xlist__new()
; Create and return a new XList.
return "\1"
endFunction

int function chat4__winrecs__xlist___indexPos(string lst, int i)
; Returns the character position in lst of the first character of the ith element.
; Returns 0 if there is no such element.
; Exception:  If i == itemCount+1, returns stringLength(lst)+1.
var
	string delim,
	int len,
	int pos,
	int result
if i < 1 then
	return 0
endIf
let delim = stringLeft(lst, 1)
let len = stringLength(lst)
if !len then
	return 0
elif len == 1 then
	return 1
endIf
let result = 0
while i
	let pos = stringContains(lst, delim)
	if !pos then
		return 0
	endIf
	let result = result +pos
	let lst = stringChopLeft(lst, pos)
	let i = i -1
endWhile
return result +1
endFunction

int function chat4__winrecs__xlist___getCharCode(string s, int whichList)
; Converts the first character in s to the corresponding code based on the indicated character table.
; If s is null, 0 is returned.
; Can't handle \000 because this seems to end JAWS strings.
; In fact, including \000 anywhere in a string causes it to appear empty!
; [DGL, 02/28/03]
let s = stringLeft(s, 1)  ; in case s contains more than one character
if !stringLength(s) then
	return 0
endIf
return stringContains(chat4__winrecs__xlist___getCharList(whichList), s)
endFunction

string function chat4__winrecs__xlist___getChar(int i, int whichList)
; Converts the given code to the corresponding character from the indicated character table.
; If i is less than 1 or above the length of the table, the null string is returned.
var string charList
if i <= 0 then
	return ""
endIf
let charList = chat4__winrecs__xlist___getCharList(whichList)
if i > stringLength(charList) then
	return ""
endIf
return substring(charList, i, 1)
endFunction

string function chat4__winrecs__xlist___getCharList(int whichList)
; Returns the requested character list.  This is to avoid duplicating huge constants in the script binary.
if whichList == chat4__winrecs__xlist___CharANSI then
	return chat4__winrecs__xlist___CharANSICharList
endIf
return ""
endFunction

int function chat4__winrecs__xlist___setDelim(string byRef lst, string ch, int dlen)
; Sets lst's delimiters to ch.
; dlen is the length of lst's current delimiter (1 except when sent via fromJAWSList).
var
	string delim
let delim = stringLeft(lst, dlen)
if util__getJFWVersion() >= 600000 && !util__isPACMate() then
	; This should be faster because (presumably) it's an in-place string edit.
	if stringLength(ch) == 1 && dlen == 1 then
		let lst = string__ref(stringReplaceChars(lst, delim, ch))
	else
		; {To/From}JAWSList() can send a multi-character delimiter here, but no proper XList will use one.
		let lst = string__ref(stringReplaceSubstrings(lst, delim, ch))
	endIf
elif dlen == 0 then
	; TODO: If we started with a null delimiter, make each character an element of the list.
	; For now, nothing is done.
else
	string__replace(lst, delim, ch, "e")
endIf
return True
endFunction

;_______________________________ End module xlist ______________________________

;--------------------- End module winrecs (from winrecs10) ---------------------

;=============================== End module chat4 ==============================


;=============================== Module roomtree ===============================
; Room tree interface for TeamTalk4Classic.
;
; Author:  Doug Lee

/*
Design info:
The following apply to the SysTreeView32 control supported by this module,
and unless otherwise noted, this module depends on these features to
work properly:

Each node has a child ID; there are no full object descendants (this
is normal for a treeview).

Child IDs are numerically contiguous at first, though they may not
start at 1.  (Most treeviews do not allow any assumptions about childID
contiguousness.)  As people join and leave the server though, holes in
the childID contiguousness develop.

A childID is never reused, even when a user leaves one server and
joins another, which means childIDs increase monotonically for as long
as the user keeps TeamTalk open.  (This fact is not used here.)

Traversing by incrementing childID seems to go in breadth-first order
(not used here).

Traversing by AccNavigate(NavDir_Down,...) goes in depth-first order
(normal for treeviews).

Spatial navigation provides correct parent/child relationships (normal
for treeviews).

The tree consists of all nodes (channels and users) immediately on server
join.  (Normal trees can leave out subnodes until the user expands them.)

Channel nodes are denoted by a parenthesized number at the end of
accName, (0) for a channel with no members.

A channel can have both member and channel descendants, but the number
in its name represents only members in that channel.

A channel's Open state flag is not reset when the channel empties even
though there may no longer be descendants.
(This module makes no use of state flags; this is just written here
because it is unusual.)

This information comes from research by DGL on 2010-05-30,
supplemented on 2010-08-30.
*/

globals
; A list of tree window handles.
	string roomtree___hwnds,
; A dictionary of storage structures by server tree window handle.
; Fields:
;	hwnd, client:  The window handle and MSAA client object for the tree.
;	nodes:  The raw list of node names at last check.
	object roomtree___trees,
; First and last childIDs in numerical order.
; Used for byAge navigation.
	int roomtree___firstID, int roomtree___lastID

int function roomtree__reportMemberships(int includeList, /*?*/ int startNode)
; Report who is where in the current TeamTalk server's channel tree.
; includeList: True to include names, False to show only counts per channel.
; TODO: This function is not fully written.
var object tree let tree = roomtree___treeRecord()
if !tree then
	sayMessage(OT_Error, roomtree___msgNoTree)
	return False
endIf
var object o let o = tree("client")
if !startNode then
	; Start on the root node.
	let startNode = 1
endIf
var string names let names = ""
var int childID let childID = o.accNavigate(7, startNode)  ; 7 is NavDir_FirstChild
var int safety let safety = 1000
var string name, int memberCount
while childID && safety
	roomtree___getNameAndCount(o.accName(childID), name, memberCount)
	if includeList then
		if memberCount == -1 then
			; A member, not a channel.
		endIf
	else
	endIf
	let childID = o.accNavigate(2, childID)  ; 2 is NavDir_Down
	let safety = safety -1
endWhile
; Safety ran out; a bad thing.
beep()
return False
endFunction

object function roomtree___treeRecord(/*?*/ handle hwnd)
; Returns the record for the given or currently focused TeamTalk's channel tree.
; Registers this tree in roomtree___trees if necessary.
; Also removes stale records for windows that no longer exist.
; Does not update the record itself from the tree.
roomtree___cleanOutStaleWindows()
var variant null
if !hwnd then
	let hwnd = roomtree__currentTreeWindow()
	if !hwnd then
		return
	endIf
endIf
var string shwnd let shwnd = intToString(hwnd)
var object tree, object o, int childID
if !roomtree___trees.exists(shwnd) then
	; We have to make a new record for this one.
	let o = getObjectFromEvent(hwnd, -4, 0, childID)  ; -4 is ObjID_Client
	if !o then
		return False
	endIf
	let roomtree___hwnds = formatString("%1|%2", roomtree___hwnds, shwnd)
	let tree = util__newCollection()
	if !roomtree___trees then
		let roomtree___trees = util__newCollection()
	endIf
	roomtree___trees.add(shwnd, tree)
	tree.add("hwnd", hwnd)
	tree.add("client", o)
	tree.add("nodes", util__newCollection())
else
	let tree = roomtree___trees(shwnd)
endIf
return tree
endFunction

int function roomtree___cleanOutStaleWindows()
; Remove from roomtree___trees any records for windows that no longer exist.
; Returns the number of records removed.
var int n let n = stringSegmentCount(roomtree___hwnds, "|")
var int nremoved let nremoved = 0
var string shwnd
while n
	let shwnd = stringSegment(roomtree___hwnds, "|", n)
	if getWindowClass(stringToHandle(shwnd)) != "SysTreeView32" then
		let nremoved = nremoved +1
		roomtree___trees.remove(shwnd)
		if !string__segmentRemove(roomtree___hwnds, "|", shwnd) then
			chat4__winrecs__log__write("Failed to remove handle " +shwnd +" from the tree list.")
		endIf
	endIf
	let n = n -1
endWhile
return nremoved
endFunction

handle function roomtree__currentTreeWindow()
; Return the handle of the channel tree for the focused TeamTalk instance,
; or return null on error.
var handle hwnd let hwnd = ttutil__mainWindow()
if !hwnd then
	return 0
endIf
let hwnd = findWindow(hwnd, "SysTreeView32", "")
if !hwnd then
	return 0
endIf
return hwnd
endFunction

int function roomtree___getNameAndCount(string s, string byRef name, int byRef count)
; Return the name and count from the given channel tree node name text.
; If the text represents a user name and not a channel (no count included),
; the returned count will be -1.
let s = stringTrimLeadingBlanks(stringTrimTrailingBlanks(s))
var int pos let pos = stringContainsFromRight(s, " ")
if !pos || substring(s, pos+1, 1) != "("
|| stringRight(s, 1) != ")" then
	let count = -1
	let name = s
	return True
endIf
let name = stringLeft(s, pos-1)
var string tmp let tmp = stringChopRight(stringChopLeft(s, pos+1), 1)
var int n let n = stringLength(tmp)
var int i let i = 1
while i <= n
	if !stringContains("0123456789", substring(tmp, i, 1)) then
		let name = s
		let count = -1
		return True
	endIf
	let i = i +1
endWhile
let count = stringToInt(tmp)
return True
endFunction

int function roomtree__isMember(object o, int childID)
; Returns True if this is a member node, as opposed to a channel.
var string name, int count
if !o.accRole(childID) then
	; This happens for a skipped childID in byAge navigation.
	return False
elif !roomtree___getNameAndCount(o.accName(childID), name, count) then
	; Unrecognized nodes are called members here.
	beep()
	return True
endIf
return (count == -1)
endFunction

int function roomtree__getMember(int childID, string which, /*?*/ int byAge, int alsoSpeak)
; Get the child ID of a member based on which and, if passed, byAge.
; "Member" means person, as opposed to channel.
; If childID is 0 or invalid, starts as appropriate from an end of the list or the current position.
; If byAge is False or not passed, uses tree order.
; If byAge is True, attempts to order members chronologically by when they last joined a channel.
; If alsoSpeak is True, speaks error messages as appropriate.
; Values of which:
;	first, last: First or last member.
;	next, prev: Next or previous member from given position.
;	nextCyclic, prevCyclic: Same but looping around the member list.
; On success, returns the childID of the member sought.
; On error, returns 0.
var object tree let tree = roomtree___treeRecord()
if !tree then
	if alsoSpeak then
		sayMessage(OT_Error, roomtree___msgNoTree)
	endIf
	return False
endIf
var object o let o = tree("client")
if byAge then
	roomtree___setFirstAndLastChildIDs(o)
endIf

; Handle next/prev requests, or if appropriate, make first/last requests out of them.
if which == "next" || which == "prev" then
	if !childID then
		let childID = o.accFocus
	endIf
	if childID then
		let childID = roomtree___navNext(o, childID, (which == "next"), byAge, True)
	endIf
	if !childID then
		; No current position or no next one,
		; so these become first/last requests in cyclic mode.
		; They also do this if no node is focused yet,
		; which happens when a server connection is first opened.
		if !stringContains(stringLower(which), "cyclic") && o.accFocus then
			if alsoSpeak then
				sayMessage(OT_Error, roomtree___msgNoMoreMembers)
			endIf
			return 0
		endIf
		if which == "next" then
			let which = "first"
		else  ; prev
			let which = "last"
		endIf
	endIf
endIf

; Handle first and last, both by tree and by age.
if which == "first" then
	; The first childID is also the oldest.
	let childID = o.accNavigate(NavDir_FirstChild, 0)
	let childID = roomtree___navNext(o, childID, True, byAge, False)
elif which == "last" then
	if byAge then
		let childID = roomtree___lastID
	else
		let childID = o.accNavigate(NavDir_LastChild, 0)
	endIf
	let childID = roomtree___navNext(o, childID, False, byAge, False)
endIf
if !childID then
	if alsoSpeak then
		sayMessage(OT_Error, roomtree___msgNoMembers)
	endIf
endIf
return childID
endFunction

int function roomtree___navNext(object o, int childID, int forward, int byAge, int skipThisOne)
; Return the next childID in the requested direction and sort order.
; If skipThisOne is True, the given node is not considered, otherwise it is.
; For byAge True, assumes roomtree___firstID and roomtree___lastID are set.
; Helper for roomtree__getMember().
var int dir
if byAge then
	if forward then
		let dir = 1
	else
		let dir = -1
	endIf
else
	if forward then
		let dir = NavDir_Next
	else
		let dir = NavDir_Previous
	endIf
endIf

while childID
	if !skipThisOne && roomtree__isMember(o, childID) then
		return childID
	endIf
	if byAge then
		let childID = childID +dir
		if childID < roomtree___firstID || childID > roomtree___lastID then
			; We hit an edge of the valid childIDs.
			let childID = 0
		endIf
	else
		let childID = o.accNavigate(dir, childID)
	endIf
	let skipThisOne = False
endWhile
return 0
endFunction

void function roomtree___setFirstAndLastChildIDs(object o)
; Sets roomtree___firstID and roomtree___lastID, used for byAge navigation.
; Helper for roomtree__getMember().
var int childID let childID = o.accNavigate(NavDir_FirstChild, 0)
; This has to be the "oldest" childID because it represents the tree root,
; which is the first childID created.
let roomtree___firstID = childID
; We know the last ID can't be less than this...
var int maxID let maxID = roomtree___firstID +o.accmaxCount -1
; but without a loop, we can't know what the highest one really is.
while childID
	if childID > maxID then
		let maxID = childID
	endIf
	let childID = o.accNavigate(NavDir_Next, childID)
endWhile
let roomtree___lastID = maxID
endFunction

int function roomtree__move(string which, int byAge)
; Move to another member node.
; Returns True on success and False on failure.
var int childID let childID = roomtree__getMember(0, which, byAge, True)
if !childID then
	return False
endIf
var object tree let tree = roomtree___treeRecord()
if !tree then
	return False
endIf
var object o let o = tree("client")
var int focusID let focusID = o.accFocus +0
if childID == focusID then
	; The indirect call avoids compilation problems before JAWS 8,
	; which are otherwise caused by the passing of a parameter.
	; The parameter avoids position info announcement like "1 of 6."
	beep()
	formatStringWithEmbeddedFunctions("<sayObjectActiveItem(0)>")
else
	; Leaving out SelFlag_TakeFocus avoids the following anomaly:
	; Focus remains, in MSAA and by window, on the right thing,
	; but Alt doesn't work, Alt+F4 destroys the tree leaving the app running,
	; and Tab and then Alt+F4 causes a busy loop terminated only by Alt+Tab.
	; Using TakeFocus+TakeSelection but calling AccSelect twice solves
	; the menu problem but causes the loop issue to occur on Alt+F4 without Tab.
	o.accSelect(SelFlag_TakeSelection, childID)
	if stringCompare(roomtree___treePath(o, childID), roomtree___treePath(o, focusID), True) != 0 then
		roomtree___sayNode()
	endIf
endIf
return True
endFunction

void function roomtree___sayNode()
; Say the path to the current tree node.
var object tree let tree = roomtree___treeRecord()
if !tree then
	return
endIf
var object o let o = tree("client")
var string path let path = roomtree___treePath(o, o.accFocus)
sayUsingVoice(VCTX_Message, path, OT_Message)
endFunction

string function roomtree___treePath(object o, int childID)
; Return the path from root to given tree node.
; Omits the given node and the top-level node.
var int safety let safety = 50
var string path let path = ""
var string name, int cnt
while safety
	let childID = o.accNavigate(NavDir_Left, childID)
	if childID && o.accNavigate(NavDir_Left, childID) then
		if !roomtree___getNameAndCount(o.accName(childID), name, cnt) then
			beep()
			return ""
		endIf
		if path then
			let path = ", " +path
		endIf
		let path = name +path
	else
		let safety = 1  ; exits loop
	endIf
	let safety = safety -1
endWhile
return path
endFunction

string function roomtree__iconStatus()
; Indicates if the user at the active cursor position is transmitting sound and/or idle.
var handle hwnd, object o, int childID
let hwnd = getCurrentWindow()
let o = getCurrentObject(childID)
if !o then
	return ""
elif !childID then
	; On top level of tree, no icons here.
	return ""
	elif !o.accRole(childID) then
		; Can happen when the tree is collapsed by a disconnect:
		; GetFocusObject() can return a stale childID.
		; [DGL, 2011-12-29, JAWS 13.0.704]
		return ""
endIf
var int left, int right, int top, int bottom
roomtree___scrollTree(hwnd, o, childID, 20)
if !MSAA__rect(o, childID, left, right, top, bottom) then
	return ""
endIf
var int x let x = left-4
var int y let y = (top+bottom) /2
var int n let n = x -getWindowLeft(getCurrentWindow())
let n = max(n, 0)
let n = min(n, 15)
var string flags let flags = ""
var int clr
var int isSending let isSending = False
var int isIdle let isIdle = False
var int foundIcon let foundIcon = False
while n
	let clr = getColorAtPoint(x, y)
	if clr == 0x0000FFL then
		let isIdle = True
	elif clr == 0x00FF00L then
		let isSending = True
	elif (clr & 0xFF) == 228 then
		let foundIcon = True
	endIf
	let x = x -1
	let n = n -1
endWhile
if isIdle then
	let flags = flags +"i"
elif foundIcon then
	let flags = flags +"o"
endIf
if isSending then
	let flags = flags +"s"
endIf
return flags
endFunction

int function roomtree___scrollTree(handle hwnd, object o, int childID, int reservedSpace)
; Scroll left/right the given tree as needed to make the following true
; regarding the MSAA rectangle's left edge (left):
;	- left is reservedSpace pixels or more right of the window's left edge.
;	- left is not right of the window's right edge.
var int left, int right, int top, int bottom
var int wleft let wleft = max(getWindowLeft(hwnd), 1)
var int wright let wright = min(getWindowRight(hwnd), screenGetWidth())
var int safety let safety = 500
var int scrollLeft, int scrollRight
while safety
	if !MSAA__rect(o, childID, left, right, top, bottom) then
		return False
	endIf
	let scrollLeft = (left -wleft < reservedSpace)
	let scrollRight = (wright -left < 1)
	if scrollLeft && scrollRight then
		return False
	endIf
	if scrollLeft then
		;beep()
		util__WMScroll(hwnd, "LineLeft")
	elif scrollRight then
		util__WMScroll(hwnd, "LineRight")
	else
		return True
	endIf
	let safety = safety -1
endWhile
return False
endFunction

;============================= End module roomtree =============================


;================================ Module ttutil ================================
; TeamTalk utility functions.
;
; Author:  Doug Lee

handle function ttutil__mainWindow()
; Return the main top-level TeamTalk window.
var handle hwnd let hwnd = getTopLevelWindow(getFocus())
if ttutil__isMainWindow(hwnd) then
	return hwnd
endIf
let hwnd = findTopLevelWindow("#32770", "")
while hwnd
	if ttutil__isMainWindow(hwnd) then
		return hwnd
	endIf
	let hwnd = getNextWindow(hwnd)
endWhile
return hwnd   ; null
endFunction

int function ttutil__isMainWindow(handle hwnd)
; Returns True if and only if hwnd is the main top-level TeamTalk window.
if getWindowClass(hwnd) == "#32770"
&& stringContains(getWindowName(hwnd), ttutil__scMainWindowName) then
	return True
endIf
return False
endFunction

;============================== End module ttutil ==============================


;================================ Module toolbar ===============================
; Module for reporting and tracking states of TeamTalk4Classic toolbar checkbox states.
;
; Author:  Doug Lee
;

globals
; Structures by top-level TeamTalk window. Fields:
;	hTop: The top-level window (same as the key for this record).
;	hToolbar, oToolbar: The handle and MSAA client object for the toolbar window.
;	sIconNames: The names of the toolbar items (none appear in MSAA).
;	sIconStates: The states of the toolbar items (see toolbar___getStates()).
	object toolbar___toolbars,
; How often toolbar__watchForChanges() watches for state changes,
; in tenths of a second.
	int toolbar___tenths,
; The schedule handle for toolbar__watchForChanges().
	int toolbar___sch

object function toolbar__record()
; Get or make the record for the currently active client's toolbar.
var handle hwnd let hwnd = getAppMainWindow(getFocus())
if !ttutil__isMainWindow(hwnd) then
	return
endIf
var object rec let rec = chat4__winrecs__record(toolbar___toolbars, hwnd)
if !rec then
	return
endIf

var handle hTop let hTop = hwnd
if !rec.exists("hTop") then
	; New record.
	let hwnd = findWindow(hwnd, "ToolbarWindow32", "")
	if !hwnd then
		return
	endIf
	var object o let o = MSAA__objectFromWindow(hwnd)
	if !o then
		return
	endIf
	var int n let n = o.accChildCount+0
	var string names
	if n == 10 then
		let names = toolbar___msgNames43
	elif n == 9 then
		let names = toolbar___msgNames42
	elif n == 8 then
		let names = toolbar___msgNames41
	else
		; TODO:  Probably should report this.
		return
	endIf
	rec.add("hToolbar", hwnd)
	rec.add("oToolbar", o)
	rec.add("sIconNames", names)
	rec.add("sIconStates", toolbar___getStates(rec))
	; Add this last as it signals record completion.
	rec.add("hTop", hTop)
endIf

return rec
endFunction

string function toolbar___getStates(object rec)
; Get the states of all toolbar items as a string.
var string states let states = ""
var object o let o = rec("oToolbar")
var int n let n = o.accChildCount+0
var int i let i = 1
while i <= n
	let states = states +"|" + intToString(o.accState(i))
	let i = i +1
endWhile
return stringChopLeft(states, 1)
endFunction

int function toolbar__checkStates()
; Check and update the states of toolbar items,
; announcing any changes as needed.
; Returns True if there were any changes and False if not.
; Meant to be called on a schedule and/or after specific key presses.
var object rec let rec = toolbar__record()
if !rec then
	return False
endIf
var string oldStates let oldStates = rec("sIconStates")
var string newStates let newStates = toolbar___getStates(rec)
if stringCompare(oldStates, newStates, True) == 0 then
	return False
endIf
var int n let n = stringSegmentCount(oldStates, "|")
if n != stringSegmentCount(newStates, "|") then
	beep()
	rec.remove("sIconStates")
	rec.add("sIconStates", newStates)
	return True
endIf
var int i let i = 1
var int oldState, int newState
while i <= n
	let oldState = stringToInt(stringSegment(oldStates, "|", i)) & State_System_Checked
	let newState = stringToInt(stringSegment(newStates, "|", i)) & State_System_Checked
	if newState != oldState then
		toolbar___speakState(newState, i)
	endIf
	let i = i +1
endWhile
rec.remove("sIconStates")
rec.add("sIconStates", newStates)
return True
endFunction

void function toolbar___speakState(int state, int i)
; Speak the given state.
var object rec let rec = toolbar__record()
var string name let name = stringSegment(rec("sIconNames"), "|", i)
if stringCompare(name, "sep", False) == 0 then
	return
endIf
if stringContains(name, "/") then
	if state then
		let name = stringSegment(name, "/", 2)
	else
		let name = stringSegment(name, "/", 1)
	endIf
elif state then
	let name = formatString(toolbar___msgActive, name)
else
	let name = formatString(toolbar___msgInactive, name)
endIf
sayUsingVoice(VCTX_Message, name, OT_Message)
endFunction

int function toolbar__watchForChanges(int tenths)
; Watch for changes in toolbar item states.
; Pass -1 to disable.
; Calls itself to do the watching.
if tenths then
	; New schedule being established.
	if toolbar___sch then
		unscheduleFunction(toolbar___sch)
		let toolbar___sch = 0
	endIf
	let toolbar___tenths = tenths
	if tenths > 0 then
		scheduleFunction("toolbar__watchForChanges", tenths)
	endIf
	return
endIf
; Internal call to do the watching.
let toolbar___sch = scheduleFunction("toolbar__watchForChanges", toolbar___tenths)
toolbar__checkStates()
endFunction

void function toolbar__sayStatuses(int verbose)
; Speak the statuses of toolbar icons.
; If verbose is False, abbreviate the results.
var object rec let rec = toolbar__record()
var string states let states = toolbar___getStates(rec)
var int n let n = stringSegmentCount(states, "|")
var int i let i = 1
var int state, string name
var string buf let buf = ""
var string sep let sep = ", "
while i <= n
	let state = stringToInt(stringSegment(states, "|", i)) & State_System_Checked
	if verbose then
		toolbar___speakState(state, i)
	else
		let name = stringSegment(rec("sIconNames"), "|", i)
		if i == 1 then
			if state then
				let buf = buf +sep +toolbar___msgConnected
			else
				let buf = buf +sep +toolbar___msgDisconnected
			endIf
		elif name != "sep" && state then
			let buf = buf +sep +name
		endIf
	endIf
	let i = i +1
endWhile
if !verbose then
	let buf = stringChopLeft(buf, stringLength(sep))
	sayMessage(OT_User_Requested_Information, buf)
endIf
endFunction

;============================== End module toolbar =============================


;================================ Module tooltip ===============================
; Functions related to the management of tooltips.
;
; Author:  Doug Lee

string function tooltip__getText(/*?*/ handle hwnd, int x, int y)
; Fire the tooltip associated with the current or given screen location if possible, and return its text.
; If x and y are not passed or passed as 0, the active cursor's location is used.
; If hwnd is not passed or passed as 0, the window at the effective point is used.
let hwnd = tooltip__getHandle(hwnd, x, y)
if !hwnd then
	return ""
endIf
return getWindowText(hwnd, Read_Everything)
endFunction

handle function tooltip__getHandle(/*?*/ handle hwnd, int x, int y)
; Create and obtain the tooltip handle associated with the current or given screen location if possible.
; If x and y are not passed or passed as 0, the active cursor's location is used.
; If hwnd is not passed or passed as 0, the window at the effective point is used.
if !x && !y then
	let x = getCursorCol()
	let y = getCursorRow()
endIf
if !hwnd then
	let hwnd = getWindowAtPoint(x, y)
endIf
var handle hwndDesk let hwndDesk = util__getDesktopWindow()
if !hwndDesk then
	return 0
endIf
var handle hwndBefore let hwndBefore = getFirstChild(hwndDesk)
if !hwndBefore then
	return 0
endIf
saveCursor()  JAWSCursor()  saveCursor()
moveTo(x, y)
winclick__send(hwnd, "hover", x, y)
pause()
var int cnt let cnt = 0
while getFirstChild(hwndDesk) == hwndBefore
	let cnt = cnt +1
	pause()
	if cnt > 10 then
		return 0
	endIf
endWhile
var handle hwndAfter let hwndAfter = getFirstChild(hwndDesk)
sayInteger(cnt)
return hwndAfter
endFunction

;============================== End module tooltip =============================


;================================== Module opt =================================
; User option code.
;
; Usage:
;	- Call opt__init() from autoStartEvent to set file and section names for options.
;	  This is optional; iff not done, getActiveConfiguration().ini and Options are used.
;	- Call opt__get() to get option values anywhere they are needed.
;	- Call opt__set() to alter option values.
;	- Call opt__cycle() from verbosity functions to handle the verbosity UI.
;
; Author:  Doug Lee

const
; A constant used to detect missing, as opposed to set-to-empty, .ini keys.
	opt___NoVal = 12345678

globals
; Information maintained by this module.
	object opt___data

void function opt__init(/*?*/ string sectname, string filename)
; Set the section and file names used by this module for storing options.
if !opt___data then
	let opt___data = util__newCollection()
endIf
opt___setVal("sectname", sectname, "Options")
opt___setVal("filename", filename, getActiveConfiguration()+".ini")
endFunction

void function opt___setVal(string name, string val, string dfl)
; Sets an internal configuration option to a given or default value.
; Helper for opt__init().
var string curval let curval = opt___data(name)
if !val && !curval then
	let val = dfl
endIf
if val then
	opt___data.remove(name)
	opt___data.add(name, val)
endIf
endFunction

variant function opt__get(string varName)
; Retrieves and returns the value of the indicated user-settable variable from the ini file for these scripts.
; This function makes sure to return integer 0 for the integer value 0, rather than returning the string "0" which shows up as True in a boolean test.
; varName: The name of the option to retrieve.
var
	string val
opt__init()
; Chop off descriptions.
let varname = stringSegment(varname, ":", 1)
let val = iniReadString(opt___data("sectname"), varname, intToString(opt___NoVal), opt___data("filename"))
; Favor integer 0 (False) rather than string "0" (True).
if stringContains(stringLeft(stringTrimLeadingBlanks(val), 1), "0") then
	return ""
endIf
if val == opt___NoVal then
	; All default values for options go here.
	if varname == SV_SpeakChatMessages then
		let val = "1"
	elif varname == SV_ChatMessageTimes then
		let val = "1"
	else
		let val = ""
	endIf
endIf
return val
endFunction

int function opt__set(string varName, variant val)
; Sets the indicated user-settable variable in TeamTalk4Classic.ini to the given value.
opt__init()
; Chop off descriptions.
let varname = stringSegment(varname, ":", 1)
return iniWriteString(opt___data("sectname"), varname, val, opt___data("filename"))
endFunction

variant function opt__cycle(string varname, string choices, int iRetCurVal)
; Used by DlgSelectFunctionToRun cycler/toggler functions to cycle/toggle variable values.
; Varname should be the option name being handled (SV_* constant).
; Choices should be a list of text values, separated by vertical bars (|).
; iRetCurVal should be the integer taken by the calling toggler/cycler function.
; The return value of this function behaves according to specifications for DlgSelectFunctionToRun functions and should be returned directly by the toggler/cycler function.
; The integer value saved in TeamTalk4Classic.ini will be the 0-based index of the chosen option.
var
	int i,
	int n
opt__init()
let i = opt__get(varname)
let n = string__segmentCount(choices, "|")
if !iRetCurVal then
	; Update it.
	let i = i +1
	if i >= n then
		let i = 0
	endIf
	opt__set(varname, i)
endIf
; i+1 rather than just i because i is 0-based and stringSegment is 1-based.
return stringSegment(choices, "|", i+1)
endFunction

;================================ End module opt ===============================


;================================ Module appOpt ================================
; User option code.

String Function jvoSpeakChatMessages (int RetCurValue)
return opt__cycle(SV_SpeakChatMessages, opt__msgOffOnChoices, RetCurValue)
EndFunction

String Function jvoChatMessageTimes (int RetCurValue)
return opt__cycle(SV_ChatMessageTimes, opt__msgOffOnChoices, RetCurValue)
EndFunction

void function appOpt___addVerbosityOption(string byRef lst, string optionData)
; Adds an option to the given verbosity (JAWS 8-) / option (JAWS 9+) list.
var
	string varAndDesc
let varAndDesc = formatString("jvo%1:%2",
	stringSegment(optionData, ":", 1),
	stringSegment(optionData, ":", 2))
if lst then
	let lst = lst +"|"
endIf
let lst = lst +varAndDesc
endFunction

Script AdjustJawsOptions ()
performScript adjustJAWSVerbosity()
endScript

Script AdjustJawsVerbosity ()
; Uses Jaws's central verbosity coding.
; Modified for skype usage by Ner
; To write more toggle functions for this, see toggle functions
; They *must* be done in the same manner.
var
	string lst

let lst = ""
appOpt___addVerbosityOption(lst, SV_SpeakChatMessages)
appOpt___addVerbosityOption(lst, SV_ChatMessageTimes)
;This next section of the script and the supporting functions were written for JAWS 9.0 and later.  This also works in older JAWS versions because the version of JFW is checked first to determine which dlg to use. [TK, 2007-10-03]
if util__getJFWVersion() > 900000 then
	OptionsTreeCore(lst)
else
	JAWSVerbosityCore(lst)
endIf
EndScript

String Function NodeHLP (string NodeName)
	if NodeName == GetActiveConfiguration () then
return appOpt__msgNodeHlp
endIf
return nodeHlp(nodeName)
EndFunction

String Function jvoSpeakChatMessagesHlp()
return stringSegment(SV_SpeakChatMessages, ":", 3)
EndFunction

String Function jvoChatMessageTimesHlp()
return stringSegment(SV_ChatMessageTimes, ":", 3)
EndFunction

;============================== End module appOpt ==============================


;================================ Module apphelp ===============================
; Automatic access to custom application help (HTM) file.
;
; The help file should be named <configName>.htm, <configName> being the result of getActiveConfiguration().
; Alternatively, the file can be named by a HelpDocument key in a Script Options section of the active configuration's .jcf file.
; As a final alternative, the name can be supplied from autoStartEvent by calling apphelp__setDocName().
;
; Author:  Doug Lee
;
; This is public code.

globals
; Custom doc name if set by a call to apphelp__setDocName().
	string apphelp___docName,
; Number of quick-succession calls to screenSensitiveHelp seen so far.
; apphelp___doHelp() also sets this to -1 to tell the screenSensitiveHelp script in this file to bypass itself and call up the chain.
	int apphelp___callCount

string function apphelp___getCustomHelpDoc()
; Gets the full path to the custom help document for the running scripts.
; Returns null if no such file is found.
var string docname let docname = apphelp___docName
if !docName then
	var string fname let fname = getActiveConfiguration() +".jcf"
	if getJFWVersion() < 600000 then
		let fname = getJAWSSettingsDirectory() +"\\" +fname
	else
		let fname = findJAWSSettingsFile(fname)
	endIf
	if stringLength(fname) && fileExists(fname) then
		let docname = iniReadString("Script Options", "HelpDocument", "", fname)
	endIf
endIf
if !docname then
	let docname = getActiveConfiguration()
	if !stringLength(docname) || docname == "default" then
		return ""
	endIf
	let docname = docname +".htm"
endIf
if getJFWVersion() < 600000 then
	let docname = getJAWSSettingsDirectory() +"\\" +docname
else
	let docname = findJAWSSettingsFile(docname)
endIf
if !fileExists(docname) then
	return ""
endIf
return docname
endFunction

void function apphelp___doHelp()
; Implements the following behavior based on the number of JAWSKey+F1 presses and the presence or absence of a custom (.htm) help file:
;	One press: Normal ScreenSensitiveHelp in 0.6 seconds.
;	Two quick presses: The custom help document opens if present, or if not, JAWS app-specific (chm) help opens.
;	Three or more quick presses: JAWS-specific (chm) help opens if present.
var
	int callCount,
		string docname
let callCount = apphelp___callCount
let apphelp___callCount = 0
if callCount != 2 then
	; 1 = screenSensitiveHelp, 3+ = app-specific help.
	; The standard screenSensitiveHelp script will use isSameScript() to decide what to do.
	let apphelp___callCount = -1
	performScript screenSensitiveHelp()
	let apphelp___callCount = 0
	return
endIf

let docname = apphelp___getCustomHelpDoc()
if docname then
	if userBufferIsActive() then
		userBufferDeactivate()
	endIf
	sayMessage(OT_Message, apphelp___msgOpening)
	run("\"" +docname +"\"")
	return
endIf

; No custom help document, so revert to standard behavior.
let apphelp___callCount = -1
performScript screenSensitiveHelp()
let apphelp___callCount = 0
endFunction

script screenSensitiveHelp()
; Intercept that makes the apphelp___doHelp() function take effect.
if apphelp___callCount == -1 then
	; apphelp___doHelp sets apphelp___callCount to -1 to indicate it wants to call the original screenSensitiveHelp script.
	performScript screenSensitiveHelp()
	return
endIf

let apphelp___callCount = apphelp___callCount +1
scheduleFunction("apphelp___doHelp", 6)
endScript

void function apphelp__setDocName(string docname)
; Set the name of the custom help doc to load.
; Call from AutoStartEvent if necessary.
let apphelp___docname = docname
endFunction

;============================== End module apphelp =============================


;================================ Module string ================================
; String manipulation functions.
;
; Author:  Doug Lee

string function string__segmentNext(string byRef s, string sep)
; Gets, returns, and removes from s the next string segment and the trailing separator if any.
; sep may be multi-character and is case sensitive.
; If sep is null, the entire string is treated as one segment.
var
	int pos,
	string s1
let pos = 0
if sep then
	let pos = stringContains(s, sep)
endIf
if !pos || !sep then
	let s1 = stringChopLeft(" " +s, 1)  ; avoids pointer duplication
	let s = ""
	return s1
endIf
let s1 = stringLeft(s, pos-1)
let s = stringChopLeft(s, pos +stringLength(sep) -1)
return s1
endFunction

int function string__segmentRemove(string byRef s, string sep, string r)
; Removes the first segment in s that exactly matches r, including case.
; Removes or leaves segment separators as appropriate.
; TODO: string__segmentRemove(s, "|", "val") is identical for s = "val" or "|val" or "val|":  the null string.
; Returns True if a segment is removed and False otherwise.
; Tested thoroughly for all segmentation and null cases 09/16/03.
var
	int slen, int seplen, int rlen, int pos
if !s || !sep || !r then
	return False
endIf

; Cases of s: (1) no r, (2) r, (3) r+sep+right, (4) left+sep+r, (5) left+sep1+r+sep2+right.
let slen = StringLength(s)
let seplen = StringLength(sep)
let rlen = stringLength(r)
let pos = stringContains(s, sep+r+sep)
if pos then
	; (5):  Most common case: r is in the middle of s somewhere, not at beginning or end.
	; This changes left+sep1+r+sep2+right to left+sep2+right.
	; pos currently points to the first character of sep1.
	let s = stringLeft(s, pos-1) +stringChopLeft(s, pos+seplen+rlen-1)
	return True
elif stringContains(s, r) == 1 then
	if slen == rlen then
		; (2) s is exactly r.
		let s = ""
		return True
	elif slen >= rlen+seplen && substring(s, rlen+1, seplen) == sep then
		; (3) s is r+sep+right (right may be null).
		let s = stringChopLeft(s, rlen+seplen)
		return True
	endIf
elif stringContains(stringRight(s, seplen+rlen), sep+r) then
	; (4):  s is left+sep+r (left may be null).
	let s = stringChopRight(s, seplen+rlen)
	return True
endif

; (1):  s does not contain r.
return False
endFunction

int function string__segmentCount(string s, string sep)
; Returns the number of segments in the given string.
; If sep is null, the result is 0.
; As of 03/15/03, the null string is defined to have zero segments (previously one).
; Any non-null string containing no separators has one segment.
; WARNING: Unlike the built-in (as of JAWS 7.0) stringSegmentCount function,
; this function counts the null string after a trailing delimiter;
; thus, string__segmentCount("a|b|", "|") is 3, where JAWS would say 2.
; Also, this function considers a multi-character sep as a multi-character delimiter,
; whereas stringSegmentCount allows each character of sep to be a delimiter.
var
	int len,
	int seplen,
	int l,
	int n
let seplen = stringLength(sep)
if util__getJFWVersion() >= 700000 && seplen == 1 then
	let n = stringSegmentCount(s, sep)
	if stringRight(s, 1) == sep then
		; JAWS does not count the null string after a trailing delimiter, but I do.
		let n = n +1
	endIf
	return n
endIf
let len = stringLength(s)
if !len then
	return 0
endIf
if !seplen then
	return 0
endIf

; A little optimization made possible by JAWS 6.0's StringReplaceSubstrings function...
if util__getJFWVersion() >= 600000 && !util__isPACMate() then
	; Count segments by removing all separators and comparing string lengths.
	let l = stringLength(string__ref(stringReplaceSubstrings(s, sep, "")))
	let l = len -l
	return (l / seplen) +1
endIf

let n = 0
let l = 0
while True
	let n = n+1
	let l = l +stringLength(stringSegment(s, sep, n)) +seplen
	if l > len then
		return n
	endIf
endWhile
; this line is never reached; even a null string will return 1 (it's a single null segment :-).
endFunction

int function string__replace(string byRef s, string r1, string r2, string whatToDo)
; Replace r1 in s with r2 if r1 is found where whatToDo indicates.
; Note:  A space is sometimes prepended to and then removed from results in this function to keep JAWS from adding
; numeric strings arithmetically, which can happen in JAWS versions at least through 3.70.47 [DGL, 12/20/2000]
; whatToDo: l=left, r=right, a=anywhere, e=everywhere.
; Add t (for Trim) if whitespace should be removed at the point where s is split.
; Add c (for ConsiderCase) if case-sensitive matching is desired.
; For compatibility with older scripts, WhatToDo can also be an integer sum of flags:
; 0=l, 1=r, 2=a, 3=e, 8=t, 16=c.
; Returns the number of changes made to s, which must be passed by reference.
; If s is null, no changes are made and 0 is returned.
; If r1 is null, the following happens based on WhatToDo:
;	"e" - r2 is inserted between each character pair in s.  All blanks are removed first if "t" is included.
;	"l" or "a" - r2 is inserted after the first character of s, after blanks around it are removed if "t" is included.
;	"r" - r2 is inserted before the last character of s, after blanks around it are removed if "t" is included.
; Author: Doug Lee
var
	string fromWhere,
	int trimEnds,
	int considerCase,
	int pos,
	int l1,
	string sa,
	string sb,
	int changed,
	int changeCount,
	int done

if !stringLength(s) then
	return 0
endIf

; Compatibility with older integer whatToDo parameter.
if stringLength(whatToDo) == 1 && stringContains(whatToDo, "0") then
	let whatToDo = "l"
	elif stringToInt(whatToDo) then
	let whatToDo = stringSegment("r|a|e|||||lt|rt|at|et|||||lc|rc|ac|ec|||||ltc|rtc|atc|etc", "|", stringToInt(whatToDo) & 31)
endIf

let whatToDo = stringLower(whatToDo)
let trimEnds = stringContains(whatToDo, "t")
let considerCase = stringContains(whatToDo, "c")
let fromWhere = whatToDo

; Special cases for null r1.
if !stringLength(r1) && fromWhere == "e" then
	if trimEnds then
		let sa = stringStripAllBlanks(s)
	else
		let sa = string__ref(s)
	endIf
	let s = ""
	let l1 = stringLength(sa)
	let pos = 1
	while pos < l1
		let s = string__cat(s, string__cat(substring(sa, pos, 1), r2))
		let pos = pos +1
	endWhile
	let s = string__cat(s, stringRight(sa, 1))
	return stringLength(sa)
elif !stringLength(r1) then
	if fromWhere == "l" || fromWhere == "a" then
		if trimEnds then
			let sb = stringTrimLeadingBlanks(s)
		else
			let sb = s +""
		endIf
		let sa = stringLeft(sb, 1)
		let sb = stringChopLeft(sb, 1)
		if trimEnds then
			let sb = stringTrimLeadingBlanks(sb)
		endIf
	else  ; fromWhere == "r"
		if trimEnds then
			let sb = stringTrimTrailingBlanks(s)
		else
			let sb = s +""
		endIf
		let sa = stringChopRight(sb, 1)
		if trimEnds then
			let sa = stringTrimTrailingBlanks(sa)
		endIf
		let sb = stringRight(sb, 1)
	endIf
	if !stringLength(sa) || !stringLength(sb) then
		return 0
	endIf
	let s = string__cat(sa, string__cat(r2, sb))
	return 1
endIf

let changeCount = 0
while !done
	let l1 = stringLength(r1)
	let changed = false
	if fromWhere == "l" then
		if !stringCompare(stringLeft(s, l1), r1, considerCase) then
			if trimEnds then
				let s = stringTrimLeadingBlanks(stringChopLeft(s, l1))
				let s = r2 + s
			else
				let s = r2 + stringChopLeft(s, l1)
			endIf
			let changed = true
		endIf
	elif fromWhere == "r" then
		if !stringCompare(stringRight(s, l1), r1, considerCase) then
			if trimEnds then
				let s = stringTrimTrailingBlanks(stringChopRight(s, l1))
				let s = s + r2
			else
				let s = stringChopRight(s, l1) + r2
			endIf
			let changed = true
		endIf
	else  ; fromWhere == "a" or "e"
		if considerCase then
			let pos = StringContains(s, r1)
		else
			let pos = StringContains(stringLower(s), stringLower(r1))
		endIf
		if pos then
			if trimEnds then
				let sa = stringTrimTrailingBlanks(stringLeft(s, pos-1))
				let sb = stringTrimLeadingBlanks(stringChopLeft(s, pos+l1-1))
				let s = " " +sa +r2 +sb
				let s = stringChopLeft(s, 1)
			else
				let s = " " +stringLeft(s, pos-1) +r2 +stringChopLeft(s, pos+l1-1)
				let s = stringChopLeft(s, 1)
			endIf
			let changed = True
		endIf
	endIf
	if fromWhere == "e" then
		let done = !changed
	else
		let done = True
	endIf
	if changed then
		let changeCount = changeCount +1
	endIf
endWhile
return changeCount
endFunction

string function string__join(string s1, string sep, string s2)
; Returns the two given strings joined with the given separator.
; If s1 or s2 is empty, the separator is not included.
if !s1 then
	return s2
elif !s2 then
	return s1
else
	return s1 +sep +s2
endIf
endFunction

string function string__ref(variant v)
; Type cast from anything to string.
; Useful for allowing object results to be included in string concatenations without a compile-time error.
return v
endFunction

string function string__cat(string s1, string s2)
; Safe string concatenator.
return stringChopLeft(" " +s1 +s2, 1)
endFunction

;============================== End module string ==============================


;================================= Module util =================================
; Miscellaneous utility functions
;
; Author:  Doug Lee

int function util__compare(variant v1, variant v2, string how)
; Comparison that returns 1 if v1 is greater than v2, 0 if equal to v2, and -1 if less than v2.
; Comparison types (values of how):
;	n - Numeric comparison (non numbers sort as 0).
;	i - Case-insensitive string comparison.
;	s (default if how is null) - Case-sensitive string comparison.
;	<funcName> - Returns funcname(v1, v2)
;	<funcCall> - Evaluates funcCall substituting %1 with "v1" and %2 with "v2."
; TODO:  Beeps and returns 0 in the following cases:
;	- String comparison (how == i or s) for JAWS versions older than 6.20.
;	- Function calls (funcName or funcCall) when v1 or v2 contains a quote (").
var
	int jver
let jver = util__getJFWVersion()
if !how then
	let how = "s"
endIf
if stringLength(how) == 1 then
	if how == "n" then
		if stringToInt(v1) == stringToInt(v2) then
			return 0
		elif stringToInt(v1) < stringToInt(v2) then
			return -1
		else
			return 1
		endIf
	elif how == "s" || how == "i" then
		if jver >= 620000 && !util__isPACMate() then
			return stringCompare(v1, v2, (how == "s"))
		else
			; TODO: Not supported and no good way to indicate it.
			beep()
			return 0
		endIf
	endIf
endIf
if stringContains(v1, "\"") || stringContains(v2, "\"") then
	; Incompatible with formatStringWithEmbeddedFunctions().
	beep()
	return 0
endIf
if stringContains(how, "(") then
	string__replace(how, "%1", "\""+string__ref(v1)+"\"", "e")
	string__replace(how, "%2", "\""+string__ref(v2)+"\"", "e")
	return stringToInt(formatStringWithEmbeddedFunctions("<" +how +">"))
endIf
return stringToInt(formatStringWithEmbeddedFunctions("<" +how +"(\"" +string__ref(v1) +"\", \"" +string__ref(v2) +"\")>"))
endFunction

int function util__abs(int n)
; Return the absolute value of n.
if n < 0 then  return 0-n
else return n
endIf
endFunction

int function util__isSpecialFocus()
; Returns True if there is a special condition for which normal focus handling should not apply.
; Currently, the conditions that are considered special are menusActive, userBufferIsActive, inHJDialog, and switchingTasks (Alt+Tab is active).
return menusActive() ||userBufferIsActive() ||inHJDialog() || util__switchingTasks()
endFunction

int function util__switchingTasks()
; Returns True if the user is currently Alt+Tabbing between tasks.

; This approach is adapted from JAWS 9.0.2169's default.jss FocusChangedEventEx function.
if getWindowClass(getFocus()) == "#32771"
&& getObjectTypeCode()== WT_LISTBOXITEM then
	return True
endIf
return False
endFunction

int function util__isPACMate()
; Returns True if this JAWS looks to be running on a PACMate.
; Technique suggested by Jonathan Avila, 2007-09-20.
var
	string sPath
let sPath = stringLower(getJAWSSettingsDirectory())
if stringContains(sPath, "\\windows\\settings\\") then
	return True
endIf
return False
endFunction

handle function util__getDesktopWindow()
; Analogous to the Windows call of the same name.
var
	handle hwnd,
	string hlist, int hcount, int i

; Switching to the invisible cursor keeps GetWindowTypeCode from returning 3 (WT_Edit) for all windows when the virtual cursor is active.
saveCursor() invisibleCursor()

; First try handles known to be the Desktop window in various versions of Windows.
; 128=95/98, 65566=NT, 65556=XP, 65546=2k.
; Thanks to Victor Tsaran for helping collect these handles. [DGL, 04/28/03]
let hlist = "128|65566|65556|65546|65536|65526|65516|65506"
let hcount = 8
let i = 1
while i <= hcount
	let hwnd = stringToHandle(stringSegment(hlist, "|", i))
	if getWindowTypeCode(hwnd) == WT_Desktop then
		return hwnd
	endIf
	let i = i +1
endWhile

; The hard way...

; Start at the focused window, or if there isn't one, from an arbitrary window on the screen.
let hwnd = getFocus()
if !hwnd then
	let hwnd = getWindowAtPoint(300, 300)  ; A valid point regardless of screen resolution
endIf
let hwnd = getFirstWindow(getTopLevelWindow(hwnd))
while hwnd
	if getWindowTypeCode(getParent(hwnd)) == WT_Desktop then
		return getParent(hwnd)
	endIf
	let hwnd = getNextWindow(hwnd)
endWhile
return 0
endFunction

string function util__callStackInfo(int nlevels, string flags, /*?*/ int skip)
; Produce a representation of the current call stack formatted according to the given parameters.
; The following shortenings are applied to each script/function call in the return value:
;	- The shared folder path is replaced with "shared."
;	- The user folder path is removed.
;	- The word "function" is removed.
;	- The word "script" is replaced with "$."
; Requires JAWS 10 or later because the JAWS 10+ getScriptCallStack() is required.
; nlevels: Number of levels from starting position up the stack to print.  1 means just current function.  0 or less means everything.
; flags: f for forward order (default is reverse), s for short (one line and no parameters).
; skip: How many calls to skip over besides this function itself, default 0.
; Returns the requested call stack information or null if the JAWS version is too old.
var
	string userFolder, string sharedFolder,
	int short, int forward,
	int stacksize
if util__getJFWVersion() < 1000000 then
	return ""
endIf
let flags = stringLower(flags)
let short = stringContains(flags, "s")
let forward = stringContains(flags, "f")
let userFolder = getJAWSSettingsDirectory()
let sharedFolder = getSharedSettingsDirectory()
var object stack let stack = util__callStack()
var int i let i = 2  ; skip this function's entry.
let i = i +skip
let stacksize = stack.count
if nlevels > 0 then
	let nlevels = min(nlevels, stacksize-i+1)
else
	let nlevels = stacksize-i+1
endIf
if nlevels <= 0 then
	return ""
endIf
var int delta let delta = 1
if forward then
	let delta = -1
	let i = i +nlevels -1
endIf
var string buf let buf = ""
while nlevels
	if buf then
		if forward then
			let buf = buf +" -> "
		else
			let buf = buf +" <- "
		endIf
	endIf
	var object  entry let entry = stack(intToString(i))
	var string fname let fname = entry("file")
	let fname = stringReplaceSubstrings(fname, sharedFolder, "")
	let fname = stringReplaceSubstrings(fname, userFolder, "user")
	if stringRight(fname, 4) == ".jsb" then
		let fname = stringChopRight(fname, 4)
	endIf
	let buf = buf +fname +" "
	if entry("type") == "script" then
		let buf = buf +"$"
	endIf
	if short then
		let buf = buf +stringSegment(entry("call"), "(", 1)
	else
		let buf = buf +string__ref(entry("call"))
	endIf
	let i = i +delta
	let nlevels = nlevels -1
endWhile
return buf
endFunction

globals object util___csRE
object function util__callStack()
; Return a dictionary of dictionaries representing the script call stack.
; Requires the VBScript.RegExp object.
; Index the result by stringified 1-based integers.
; The most recent call is item 1.
; Structure of an item:
;	file: The file from which the script/function came, either "builtin" or a full path.
;	type: The type ("script" or "function", lower case).
;	call: The full call with parameters.
if !util___csRE then
	let util___csRE = util__createObject("VBScript.RegExp")
	if !util___csRE then
		return
	endIf
	let util___csRE.Global = -1
	let util___csRE.IgnoreCase = 0
	let util___csRE.Pattern = "\10([^\10\t]*)\t(script |function )"
endIf
var string sStack let sStack = "\10" +getScriptCallStack()
var object matches let matches = util___csRE.execute(sStack)
var object match
var int n let n = matches.count
var int i let i = 1  ; Skip this function itself.
var string tmp, int pos1, int pos2
var object stack let stack = util__newCollection()
var object entry
while i < n
	let entry = util__newCollection()
	stack.add(intToString(stack.count), entry)
	let match = matches(i)
	; Skip the \10 here.
	; Also switch from 0-base to 1-base.
	let pos1 = match.firstIndex +2
	if i == n-1 then
		let pos2 = stringLength(sStack) +1
	else
		let pos2 = matches(i+1).firstIndex +1
	endIf
	let tmp = substring(sStack, pos1, pos2-pos1)
	entry.add("file", string__segmentNext(tmp, "\t"))
	entry.add("type", string__segmentNext(tmp, " "))
	entry.add("call", tmp)
	let i = i +1
endWhile
return stack
endFunction

globals
	int util___JAWSVer
int function util__getJFWVersion()
; GetJFWVersion with caching to make it faster.
; Also includes the build number even when GetJFWVersion() does not.
var
	string js
if !util___JAWSVer then
	let util___JAWSVer = getJFWVersion()
	let js = intToString(util___JAWSVer)
	if stringRight(js, 3) == "000" then
		let js = util__getJAWSInfo()
		let js = stringSegment(js, "|", 1)
		let js = stringSegment(js, ".", 3)
		let util___JAWSVer = util___JAWSVer +stringToInt(js)
	endIf
endIf
return util___JAWSVer
endFunction

string function util__getJAWSInfo()
; Returns a list of fields, separated by "|", as follows:
; 1.  The JAWS version number from the JFWUI2 window, in the form [M]M.mm.bbb; example, 5.00.584.
; 2.  The JAWS serial number from the JFWUI2 window.
; 3.  The JAWS serial number from GetJFWSerialNumber().
; The build number returned by this function is accurate, whereas it is "000" in GetJFWVersion in older JAWS versions.
var
	handle hwnd0, handle hwnd,
	int pos,
	string sn, string ver

; Find the JAWS dialog, which exists even when invisible or when JAWS is on the system tray.
let hwnd0 = getFirstChild(findTopLevelWindow("JFWUI2", "JAWS"))
if !hwnd0 then
	return ""
endIf

; Get the serial number from the status bar's window name.
let hwnd = getNextWindow(hwnd0)
let sn = stringLower(getWindowName(hwnd))
let pos = stringContains(sn, "ber")
if pos then
	let sn = intToString(stringToInt(substring(sn, pos+4, 10)))
else
	let sn = ""
endIf

; Find the static containing the version number and get it.
let hwnd = getFirstChild(hwnd0)
let ver = ""
while hwnd && !ver
	let ver = stringTrimTrailingBlanks(getWindowName(hwnd))
	let pos = stringContainsFromRight(ver, " ")
	if pos then
		let ver = stringChopLeft(ver, pos)
	else
		let ver = ""
	endIf
	let hwnd = getNextWindow(hwnd)
endWhile

if !sn && !ver then
	return ""
else
	return ver +"|" +sn +"|" +getJFWSerialNumber()
endIf
endFunction

object function util__createObject(string objname)
; A CreateObject wrapper that avoids causing 5-7 second delays in DOS boxes when called if JAWS is running as a service.
; As of June 11, 2011, uses the CoCreateInstance method of creating an object
; unless that doesn't work, whereupon the GetObjectHelper method is used.
; Previously the CoCreateInstance method was only used if JAWS was running as a service.
var object o let o = createObjectEx(objname, False)
if !o then
	let o = createObject(objname)
endIf
return o
endFunction

int function util__WMScroll(handle hwnd, string cmd)
; Scroll the given window or get info about its scroll position according to cmd.
; Commands allowed (18):
;	LineLeft/Right/Up/Down, PageLeft/Right/Up/Down.
;	Left/Top/Right/Bottom.
;	H/VThumbPosition/Track.
;	H/VEndScroll.
var
	int msg, int wParam
; Translate cmd in place into something more directly parseable - a two-character code:
;	h/v:  Horizontal or vertical operation.
;	-+<>^$pte:
;		-+<>:  Line or page forward/back.
;		^$:  Fully left/top or right/bottom.
;		pte:  ThumbPosition, ThumbTrack, or EndScroll.
; This system avoids up to 18 string comparisons and guarantees that cmd is valid.
if !util__stringTran(cmd, False,
		" LineLeft LineRight LineUp LineDown PageLeft PageRight PageUp PageDown Left Top Right Bottom HThumbPosition VThumbPosition HThumbTrack VThumbTrack HEndScroll VEndScroll",
		" h- h+ v- v+ h< h> v< v> h^ v^ h$ v$ hp vp ht vt he ve"
) then
	return False
endIf

; Now set msg and wParam.
if cmd == "h" then
	let msg = 0x0114  ; WM_HScroll
else
	let msg = 0x0115  ; WM_VScroll
endIf
; This produces an SB_* constant from winuser.h.
let wParam = stringContains("-+<>pt^$e", stringRight(cmd, 1)) -1
return sendMessage(hwnd, msg, wParam, 0)
endFunction

int function util__stringTran(string byRef val0, int caseSensitive, string s1, string s2)
; Translate val in place by finding it in s1 and replacing it with the corresponding s2 value.
; Returns True on success and False on failure.
; TODO: This function probably belongs in the String module.
var
	string val,
	string sep1, string sep2,
	int idx
let val = val0
if !caseSensitive then
	let val = stringLower(val)
	let s1 = stringLower(s1)
endIf
let sep1 = stringLeft(s1, 1)
let sep2 = stringLeft(s2, 1)
let idx = stringSegmentIndex(s1, sep1, val, True)
if !idx || idx > stringSegmentCount(s2, sep2) then
	return False
endIf
let val0 = stringSegment(s2, sep2, idx)
return True
endFunction

object function util__newCollection()
; Return something permitting storage of a collection of data items.
; This is here to simplify migration from JAWS 10 and old 11 to JAWS 11 update 1 and later.
var object coll let coll = createObjectEx("Scripting.Dictionary", False)
let coll.CompareMode = 1  ; TextCompare
return coll
endFunction

variant function util__vcast(variant v)
; Cast any type to variant.
return v
endFunction

;=============================== End module util ===============================


;=============================== Module winclick ===============================
; Support for sending mouse clicks of various sorts directly to a window.
; This permits clicking on obscured windows
; or clicking without checking for what is in the foreground.
;
; Author:  Doug Lee, with credit to Jim Bauer for the idea and some initial research.
;
; TODO:
;	- No support for mouse wheel events.
;	- Alt+clicks not supported.
;	- Non-client clicks not supported, such as for scrollbar elements.


int function winclick__send(handle hwnd, string which, /*?*/ int x, int y)
; Send a click to the given window using a sendMessage() call.
; hwnd: The window to which to send an event.
; which: Which event and keys to include.  Click specifiers:
;	Left, Right, or Middle:  Which button to use for a click.
;	dbl (optional):  Include if you want a double click.
;	Click (optional):  Not required.
;	Hover: Use in place of the above to cause a hover event instead of a click.
;	Leave: Use in place of the above to cause a leave event.
; Examples: leftDblClick, leftClick, leftDbl, left, hover.
; Key specifiers (include zero or more):
;	ctrl or control: A control key is down.
;	shift: A shift key is down.
; x: The x coordinate to click or hover over (optional).
; y: The y coordinate to click or hover over (optional).
; If x and y are not passed or are 0,
; the top left corner of the Window's client area is used.

; Returns the sendMessage() result.
return winclick___doClick(True, hwnd, which, x, y)
endFunction

void function winclick__post(handle hwnd, string which, /*?*/ int x, int y)
; Send a click to the given window using a sendMessage() call.
; hwnd: The window to which to send an event.
; which: Which event and keys to include.  Click specifiers:
;	Left, Right, or Middle:  Which button to use for a click.
;	dbl (optional):  Include if you want a double click.
;	Click (optional):  Not required.
;	Hover: Use in place of the above to cause a hover event instead of a click.
;	Leave: Use in place of the above to cause a leave event.
; Examples: leftDblClick, leftClick, leftDbl, left, hover.
; Key specifiers (include zero or more):
;	ctrl or control: A control key is down.
;	shift: A shift key is down.
; x: The x coordinate to click or hover over (optional).
; y: The y coordinate to click or hover over (optional).
; If x and y are not passed or are 0,
; the top left corner of the Window's client area is used.

winclick___doClick(False, hwnd, which, x, y)
endFunction

int function winclick___doClick(int useSend, handle hwnd, string which, /*?*/ int x, int y)
; Send a click to the given window using a sendMessage() or postMessage call.
; See winclick__sendClick() or winclick__postClick() for details on usage.
; Returns the sendMessage() result if sendMessage() is used, or nothing if postMessage() is used.
; Fine print: This function sends two messages, ; one for buttonDown and one for buttonUp.
; The sendMessage result comes from the first of these calls when sendMessage is used.
; Only one message is sent for a hover or leave.
let which = stringLower(which)
var int msgDown, int msgUp
var int wParam let wParam = 0
if stringContains(which, "hover") then
	let msgDown = 0x02A1  ; WM_MOUSEHOVER
	let msgUp = 0  ; skip that one.
elif stringContains(which, "leave") then
	let msgDown = 0x02A3  ; WM_MOUSELEAVE
	let msgUp = 0  ; skip that one.
elif stringContains(which, "left") then
	if stringContains(which, "dbl") then
		let msgDown = 0x0203  ; WM_LBUTTONDBLCLK
	else
		let msgDown = 0x0201  ; WM_LBUTTONDOWN
	endIf
	let msgUp = 0x0202  ; WM_LBUTTONUP
	let wParam = wParam | 1  ; MK_LBUTTON
elif stringContains(which, "right") then
	if stringContains(which, "dbl") then
		let msgDown = 0x0206  ; WM_RBUTTONDBLCLK
	else
		let msgDown = 0x0204  ; WM_RBUTTONDOWN
	endIf
	let msgUp = 0x0205  ; WM_RBUTTONUP
	let wParam = wParam | 2  ; MK_RBUTTON
elif stringContains(which, "middle") then
	if stringContains(which, "dbl") then
		let msgDown = 0x0209  ; WM_MBUTTONDBLCLK
	else
		let msgDown = 0x0207  ; WM_MBUTTONDOWN
	endIf
	let msgUp = 0x0208  ; WM_MBUTTONUP
	let wParam = wParam | 0x10  ; MK_MBUTTON
else
	beep()
	return False
endIf
if stringContains(which, "shift") then
	let wParam = wParam | 4  ; MK_SHIFT
endIf
if stringContains(which, "ctrl") || stringContains(which, "control") then
	let wParam = wParam | 8  ; MK_CONTROL
endIf
var int lParam let lParam = 0
if x || y then
	if !msgUp then
		saveCursor()  JAWSCursor()  saveCursor()
		moveTo(x, y)
	endIf
	var int left, int right, int top, int bottom
	; -4 is ObjID_Client
	if MSAA__windowRect(hwnd, -4, left, right, top, bottom) then
		let x = x -left
		let y = y -top
	else
		let x = x -getWindowLeft(hwnd)
		let y = y -getWindowTop(hwnd)
	endIf
	let lParam = makeLong(x, y)
endIf
var int result
if useSend then
	let result = sendMessage(hwnd, msgDown, wParam, lParam)
	if msgUp then
		sendMessage(hwnd, msgUp, 0, lParam)
	endIf
else
	postMessage(hwnd, msgDown, wParam, lParam)
	if msgUp then
		postMessage(hwnd, msgUp, 0, lParam)
	endIf
endIf
if !msgUp then
	pause()
endIf
return result
endFunction

;============================= End module winclick =============================


;================================= Module MSAA =================================
; Miscellaneous functions for MSAA access
;
; Author:  Doug Lee

int function MSAA__windowState(handle hwnd, /*?*/ int childID, int objID)
; Get the MSAA state flags for the given window.
return MSAA__windowProp(hwnd, "state", childID, objID)
endFunction

int function MSAA__windowRect(handle hwnd, int objID, int byRef left, int byRef right, int byRef top, int byRef bottom)
; Return by reference the rectangle for the given object ID in the given window.
; Returns True on success and False on failure.
let left = 0
let right = 0
let top = 0
let bottom = 0
var object o, int childID
let o = getObjectFromEvent(hwnd, objID, 0, childID)
if !o then
	return False
endIf
if !MSAA__rect(o, 0, left, right, top, bottom) then
	return False
endIf
return True
endFunction

variant function MSAA__windowProp(handle hwnd, string whichProp, /*?*/ int childID, int objID)
; Returns any one of several possible MSAA properties for the given window.
; Supported properties:
;	Name, Role, State, Value.
;	Description, KeyboardShortcut, Help, ChildCount.
var
	variant result,
	object o
if whichProp == "childCount" || whichProp == "role" || whichProp == "state" then
	let result = util__vcast(0)
else
	let result = util__vcast("")
endIf
if !hwnd then
	return result
endIf
let o = MSAA__objectFromWindow(hwnd, objID)
if !o then
	return result
endIf
let childID = childID +0
if whichProp == "name" then
	return o.accName(childID)
elif whichProp == "state" then
	return o.accState(childID)
elif whichProp == "role" then
	return o.accRole(childID)
elif whichProp == "value" then
	return o.accValue(childID)
elif whichProp == "description" then
	return o.accDescription(childID)
elif whichProp == "help" then
	return o.accHelp(childID)
elif whichProp == "keyboardShortcut" then
	return o.accKeyboardShortcut(childID)
elif whichProp == "childCount" then
	if childID then
		return o.accChild(childID).accChildCount
	else
		return o.accChildCount
	endIf
endIf
return result
endFunction

int function MSAA__rect(object o, int childID, int byRef left, int byRef right, int byRef top, int byRef bottom)
; Get the bounding rectangle for the given MSAA object and childID.
; The rectangle is returned 1-based, as JAWS likes its rectangles.
; MSAA provides 0-based left/top/width/height.
if !o then
	let left = 0
	let right = 0
	let top = 0
	let bottom = 0
	return False
endIf
if !childID then
	let childID = 0
endIf
o.accLocation(intRef(left), intRef(top), intRef(right), intRef(bottom), childID)
if !left && !right && !top && !bottom then
	return False
endIf
; but we want 1-based left/right/top/bottom.
let left = left +1
let top = top +1
let right = left +right
let bottom = top +bottom
return True
endFunction

globals
	handle MSAA___lastHwnd,
	int MSAA___lastObjID,
	object MSAA___lastObj

object function MSAA__objectFromWindow(handle hwnd, /*?*/ int objID)
; Return the requested (default ObjID_Client) MSAA object from the given window.
if !stringLength(util__vcast(objID)) then
	let objID = -4  ; ObjID_Client
endIf
if hwnd == MSAA___lastHwnd && objID == MSAA___lastObjID && MSAA___lastObj then
	return MSAA___lastObj
endIf
var object o, int childID
let o = getObjectFromEvent(hwnd, objID, 0, childID)
if !o then
	return o
endIf
let MSAA___lastHwnd = hwnd
let MSAA___lastObjID = objID
let MSAA___lastObj = o
return o
endFunction

;=============================== End module MSAA ===============================

; the additional scenario
; The Author - Mihail Gordeev

; conservation is audio to actuate and switch off
Script StoreAudio  ()
Var
Handle hWnd

TypeKey ("Insert+3")
TypeKey ("Ctrl+Alt+A")
Pause ()
Let hWnd = FindDescendantWindow (GetParent (GetFocus ()), 1364)
If (hWnd) Then
SpeechOff ()
PostMessage (hWnd, BM_CLICK)
Let hWnd = FindDescendantWindow (GetParent (GetFocus ()), 1)
Pause ()
PostMessage (hWnd, WM_LBUTTONDOWN, 0)
PostMessage (hWnd, WM_LBUTTONUP, 0)
;Pause ()
PlaySound (GetSharedSettingsDirectory ()+"\\SOUNDS\\BlindsDownOnly.wav")
SpeechOn ()
Say ("", ot_message)
Return true
EndIf
Say (" ", ot_message)
EndScript

; Say a file name
Script SayNameFile ()
var
Handle hWnd

Let hWnd = GetFocus ()
If GetWindowClass (hWnd) == "Button" Then
Let hWnd = GetNextWindow (GetFirstWindow (hWnd))
Say (" " + GetWindowText (hWnd, false), ot_message)
endif
EndScript

; Say a size of the file
Script SaySizeFile ()
var
Handle hWnd,
String symbol,
String s

Let hWnd = GetFocus ()
if GetWindowClass (hWnd) == "Button" then
Let hWnd = GetNextWindow (GetNextWindow (GetNextWindow (GetFirstWindow (hWnd))))
Let s = GetWindowText (hWnd, false)
Let symbol = StringLeft (StringSegment (s, cScSpace, -1), 1)
Say ("", ot_message)
If symbol == "B" Then
SayLoading (StringSegment (s, cScSpace, 1))
ElIf symbol == "K" Then
SayLoading (IntToString (StringToInt (StringSegment (s, cScSpace, 1)) * 1024))
Else
Say (s, ot_message)
endif
EndIf
EndScript

; Say a size of the jumped file from the general size of the file
Script SaySizeLoading ()
var
Handle hWnd,
String s

Let hWnd = GetFocus ()
if GetWindowClass (hWnd) == "Button" then
Let hWnd = GetPriorWindow (GetPriorWindow (GetPriorWindow (GetLastWindow (hWnd))))
Let s = StringSegment (GetWindowText (hWnd, false), cScSpace, 1)
Say (" ", ot_message)
SayLoading (StringSegment (s, "/", 1))
Say ("", ot_message)
SayLoading (StringSegment (s, "/", 2))
endif
EndScript

; Say a percent of the jumped file
Script SayPercentLoading ()
var
Handle hWnd,
String s

Let hWnd = GetFocus ()
if GetWindowClass (hWnd) == "Button" then
Let hWnd = GetPriorWindow (GetPriorWindow (GetPriorWindow (GetLastWindow (hWnd))))
Let s = StringSegment (StringSegment (GetWindowText (hWnd, false), "-", 2), ".", 1) + "%"
Say (" " +s, ot_message)
endif
EndScript

; Say a velocity of  file
Script SaySpeedLoading ()
var
Handle hWnd,
String s

Let hWnd = GetFocus ()
if GetWindowClass (hWnd) == "Button" then
Let hWnd = GetNextWindow (GetNextWindow (GetNextWindow (GetNextWindow (GetFirstWindow (hWnd)))))
Let s = GetWindowText (hWnd, false)
Say (" " + StringSegment (s, ",", 1), ot_message)
endif
EndScript

	Void Function SayLoading (String s)
Var
String ost,
Int n,
Int k,
Int r

Let n = StringToInt (s)
If n >= 1048576 Then
Let k = n / 1048576
Let r = n - k * 1048576
Let r = r * 10 / 1048576
Let ost = " "
ElIf n >= 1024 Then
Let k = n / 1024
Let r = n - k * 1024
Let r = r * 10 / 1024
Let ost = ""
ElIf n< 1024 Then
Let k = n
Let ost = ""
EndIf
If r >= 5 Then
Let k = k + 1
EndIf
SayInteger (k)
Say (ost, ot_message)
EndFunction

Script MinimizeWindow ()
;TypeCurrentScriptKey ()
TypeKey ("Alt+M")
PlaySound (GetSharedSettingsDirectory ()+"\\sounds\\CameraShutter1.wav")
Say ("  ", ot_message)
EndScript

Script SeTTree ()
Var
Handle hWnd

let hWnd = FindWindow (GetAppMainWindow (GetCurrentWindow ()), "SysTreeView32", "")
If (hWnd) Then
PostMessage (hWnd, WM_LBUTTONDOWN, 0)
PostMessage (hWnd, WM_LBUTTONUP, 0)
EndIf
EndScript

Script SetChatFiles ()
Var
Handle hWnd

let hWnd = FindWindow (GetAppMainWindow (GetCurrentWindow ()), "SysTabControl32", "")
If (hWnd) Then
If FindString (hWnd, StringSegment (vk, ";", StringContains ("cf", StringLower (StringRight (GetCurrentScriptKeyName (), 1)))), 0, 0) Then
Pause ()
LeftMouseButton ()
PCCursor ()
SetFocus (hWnd)
EndIf
EndIf
EndScript

Script LockedAllRoom ()
Var
Handle hWnd

Let hWnd = GetFocus ()
If GetWindowClass (hWnd) == "SysTreeView32" Then
If Not StringIsBlank (gRoom) && FindGraphic (hWnd, gRoom, S_TOP, S_UNRESTRICTED) Then
RoutePCToJaws ()
TypeKey ("LeftArrow")
EndIf
PCCursor ()
Return
EndIf
Say ("  ", ot_message)
		EndScript

; Say who speaks
Script WhoIsTalk ()
Var
Handle hWnd

Let hWnd = GetFocus ()
If GetWindowClass (hWnd) == "SysTreeView32" Then
Pause ()
If Not StringIsBlank (speaks) && FindGraphic (hWnd, speaks, S_TOP, S_UNRESTRICTED) Then
Pause ()
SayTextBetween (GetWindowLeft (hWnd), GetWindowRight (hWnd))
RoutePCToJaws ()
Say (" ", ot_message)
EndIf
PCCursor ()
Else
Say ("  ", ot_message)
EndIf
EndScript

Script AllIsTalks ()
Var
Handle hWnd

Let hWnd = GetFocus ()
If GetWindowClass (hWnd) == "SysTreeView32" Then
Let nLeftDrive = GetWindowLeft (hWnd)
Let nRightDrive = GetWindowRight (hWnd)
Let talks = ""
GraphicsEnumerate (hWnd, "EnumFunc") 
Say (talks, ot_message)
PCCursor ()
Else
Say ("  ", ot_message)
EndIf
EndScript

Int Function EnumFunc (int nX1, int nY1, int nX2, int nY2)
InvisibleCursor ()
;       
;   , 
;     
MoveTo ((nX1 +nX2)/2, (nY1 +nY2)/2)
If GetWord () == speaks Then
;SayTextBetween (nLeftDrive, nRightDrive)
Let talks = talks + GetTextBetween (nLeftDrive, nRightDrive) + ". "
EndIf
return TRUE
EndFunction

Script SayNewMessage ()
If (IsSameScript ()) then
CopyToClipboard (inform)
SayMessage (ot_user_requested_information,"  ")
Return
EndIf
Say (inform, ot_message)
EndScript

String Function GetBottomLineOfWindow(handle hWnd)
var
 int iLeft,
 int iTop,
 int iRight,
 int iBottom

GetWindowRect(hWnd,iLeft,iRight,iTop,iBottom)
let sWinText =
GetTextInRect(iLeft,iTop,iRight,iBottom,0,IgnoreColor,IgnoreColor,1,0)
Return sWinText
EndFunction

Script ChangeStateWindow ()
Var
String state

If iWnd == 61488 Then
Let iWnd = 61728
Let state = " "
ElIf iWnd == 61728 Then
Let iWnd = 61488
Let state = " "
EndIf
SendMessage (GetAppMainWindow (GetCurrentWindow ()), 274, iWnd, 0)
Say (state, ot_message)
EndScript

Script ToggleVerbosity ()
TypeKey ("Control+Alt+s")
Let nFlagEvent = 1
SpeechOff ()
TypeKey ("F4")
EndScript


Script ToggleVoiceActivation ()
;TypeKey ("Control+Shift+a")
TypeCurrentScriptKey ()
Let nFlagVoice = 1
SpeechOff ()
TypeKey ("F4")
EndScript

Script TogglePushToTalk ()
;TypeKey ("Control+Shift+t")
TypeCurrentScriptKey ()
Let nFlagPush = 1
SpeechOff ()
TypeKey ("F4")
EndScript

Script UpdatedChannel ()
Let nUpdatedChannel = 1
TypeKey ("Shift+F7")
EndScript

Script RunF2 ()
Let nConnect = 1
TypeKey ("F2")
EndScript

Script SayPing ()
Say (". " + ping, ot_message)
EndScript

Void Function SayHighLightedText (handle hwnd, string buffer)

If GetWindowName (hWnd) == " " && nConnect == 1 Then
Return true
EndIf

If GetWindowClass (hWnd) == "Listbox" && GetControlId (hWnd) == 1154 Then
Return true
EndIf
SayHighLightedText (hwnd, buffer)
EndFunction  

Void Function SayNonHighLightedText (handle hwnd, string buffer)
Var
String s,
String str,
Int n,
Int i

If GetWindowClass (hWnd) == "msctls_statusbar32" && GetControlId (hWnd) == 59393 Then
If StringContains (buffer, "Disconnected") Then
TypeKey ("F2")
StopSpeech ()
EndIf

If StringContains (buffer, "Finish") || StringContains (buffer, "Start") Then
Let inform = buffer
Say (buffer, ot_message)
EndIf
Let s = StringSegment (buffer, ":", -2)		
Let s = StringLeft (s, StringContains (s, "") -1)
If StringCompare (s, ping, false) && !StringContains (s, "KB")  && s != "" Then
Let ping = s
EndIf
Return
EndIf

If GetWindowClass (hWnd) == "ListBox" && GetControlId (hWnd) == 1154 && nConnect == 1 Then
PostMessage (hWnd, WM_LBUTTONDOWN, 0)
PostMessage (hWnd, WM_LBUTTONUP, 0)
Let nConnect = 0
Return true
EndIf

If GetWindowClass (hWnd) == "RichEdit20W" && GetWindowName (hWnd) == "" Then
GetBottomLineOfWindow(hWnd)
Let n = StringSegmentCount (sWinText, "\n")
Let s = StringSegment (sWinText, "\n", 1) + " . "
Let i = 1
While i <= n 
Let i = i + 1
Let str = StringSegment (sWinText, "\n", i) +"'"
If !StringContains (str, ":'") Then
Let s = s + str
Else
let i = n
EndIf
EndWhile
Pause ()
Let s = StringReplaceSubstrings (s, "'", cScSpace)
If StringCompare (s, inform, false) Then
Say (s, ot_message)
Let inform = s
EndIf
Return
EndIf
SayNonHighLightedText (hwnd, buffer)
EndFunction 

int Function HandleCustomWindows (handle hWnd)
Var
Handle hWnd1

If GetWindowClass (hWnd) == "Edit" && GetControlId (hWnd) == 1031 && nUpdatedChannel == 1 Then
Let hWnd1 = FindDescendantWindow (GetForegroundWindow (), 1303)
PostMessage (hWnd1, BM_CLICK)
Let nVerb = SendMessage (hWnd1, 240)
Let hWnd1 = FindDescendantWindow (GetParent (GetFocus ()), 1)
PostMessage (hWnd1, WM_LBUTTONDOWN, 0)
PostMessage (hWnd1, WM_LBUTTONUP, 0)
Pause ()
Say ("  . " + StringSegment (verb, ";", 2 - nVerb), ot_message)
Let nUpdatedChannel = 0
Return true
EndIf

If GetWindowClass (hWnd) == "Edit" && GetControlId (hWnd) == 1052 && nFlagEvent == 1 Then
MoveToWindow (FindDescendantWindow (GetForegroundWindow (), 12320))
LeftMouseButton ()
PCCursor ()
EndIf
If GetWindowClass (hWnd) == "Edit" && GetControlId (hWnd) == 1100 && nFlagEvent == 1 Then
Let nVerb = SendMessage (FindDescendantWindow (GetForegroundWindow (), 1023), 240)
PostMessage (FindDescendantWindow (GetForegroundWindow (), 2), BM_CLICK)
Let nFlagEvent = 0
SpeechOn ()
Pause ()
Say (". " + StringSegment (verb, ";", nVerb + 1), ot_message)
Return true
EndIf

If GetWindowClass (hWnd) == "Edit" && GetControlId (hWnd) == 1052 && nFlagVoice == 1 Then
Let nVerb = SendMessage (FindDescendantWindow (GetForegroundWindow (), 1056), 240)
PostMessage (FindDescendantWindow (GetForegroundWindow (), 2), BM_CLICK)
Let nFlagVoice = 0
SpeechOn ()
Pause ()
Say (" . " + StringSegment (verb, ";", nVerb + 1), ot_message)
Return true
EndIf

If GetWindowClass (hWnd) == "Edit" && GetControlId (hWnd) == 1052 && nFlagPush == 1 Then
Let nVerb = SendMessage (FindDescendantWindow (GetForegroundWindow (), 1054), 240)
PostMessage (FindDescendantWindow (GetForegroundWindow (), 2), BM_CLICK)
Let nFlagPush = 0
SpeechOn ()
Pause ()
Say (",  . " + StringSegment (verb, ";", nVerb + 1), ot_message)
Return true
EndIf

If GetWindowClass (hWnd) == "Button" && (GetWindowName (GetParent (hWnd)) == "  " || GetWindowName (GetParent (hWnd)) == "  " || GetWindowName (GetParent (hWnd)) == "    ") && GetWindowClass (GetPriorWindow (hWnd)) == "Edit" Then
Say (" ", ot_message)
EndIf
If GetWindowClass (hWnd) == "Button" && (GetWindowName (GetPriorWindow (hWnd)) == "  " || GetWindowName (GetPriorWindow (hWnd)) == "Folder for audio files") Then
Say (" ", ot_message)
EndIf
If GetWindowClass (hWnd) == "Button" &&( (GetWindowName (hWnd) == " " || GetWindowName (hWnd) == "Clear") 
|| GetWindowName (GetPriorWindow (GetPriorWindow (hWnd))) == "") Then
Say (GetWindowName (GetPriorWindow (GetPriorWindow (hWnd))), ot_message)
EndIf

If GetWindowClass (hWnd) == "Button" && (GetWindowName (hWnd) == "" || GetWindowName (hWnd) == "OK") && GetWindowClass (GetParent (hWnd)) == "#32770" && (GetWindowName (GetParent (hWnd)) == "TeamTalk5Classic" || GetWindowName (GetParent (hWnd)) == " ") Then
Let inform = StringChopRight (GetWindowText (GetParent (hWnd), false), 3)
Pause ()
EnterKey ()
Return true
EndIf

If GetWindowName (hWnd) == " " && nConnect == 1 Then
Return true
Endif

EndFunction 

int Function HandleCustomRealWindows (handle hWnd)
If GetWindowClass (hWnd) == "#32770" && GetWindowName (hWnd) == " " && nUpdatedChannel == 1 Then
Return True
EndIf
EndFunction

; hot keys scenario
Script hotkeyhelp ()
If UserBufferIsActive () then
UserBufferDeactivate ()
EndIf
SayFormattedMessage (OT_USER_BUFFER, msghelp, msghelp)
EndScript

; Insert smiles
Script InsertSmiles ()
var
string SmilesName,
string smile,
String SmilesIniFile,
int choice
let SmilesIniFile =GetJAWSSettingsDirectory () + "\\TeamTalk4Classic.ini"
let SmilesName = IniReadSectionKeys ("SmilesName", SmilesIniFile)
let choice = DlgSelectItemInList (SmilesName, " ", 0, 1)
if choice then
let smilesName = StringSegment (SmilesName, "|", choice)
Let Smile = IniReadString ("SmilesName", SmilesName, "", SmilesIniFile) 
TypeString (" " +smile)
endif
EndScript

Script ProgramKey ()
TypeCurrentScriptKey ()
EndScript

Script ToggleSilenceChannel ()
Var
Int nSilence

TypeKey ("Control+M")
SpeechOff ()
ActivateMenuBar ()
Delay (2)
TypeString ("")
EnterKey ()
PriorLine ()
PriorLine ()
Pause ()
Let nSilence = GetObjectStateCode ()
Pause ()
ActivateMenuBar ()
Pause ()
SpeechOn ()
Say (". " + StringSegment (verb, ";", nSilence + 1), ot_message)
EndScript

Script ToggleVoiceUser ()
Var
Int nSilence,
String s

TypeKey ("Control+Shift+M")
Let s = GetToEndOfLine ()
SpeechOff ()
ActivateMenuBar ()
Delay (2)
TypeString ("")
EnterKey ()
TypeString ("")
EnterKey ()
Pause ()
Let nSilence = GetObjectStateCode ()
Pause ()
ActivateMenuBar ()
Pause ()
SpeechOn ()
Say (s, ot_message)
If nSilence != 1 && nSilence !=0 Then
Say ("", ot_message)
Return
EndIf
Say (". " + StringSegment (verb, ";", 2 - nSilence), ot_message)
If nSilence == 1 Then
Let VoiceNone = VoiceNone + s + ";"
ElIf nSilence == 0 Then
Let VoiceNone = StringReplaceSubstrings (VoiceNone, s + ";", "")
EndIf
EndScript

Script ToggleMediaFile ()
Var
Int nSilence

;TypeKey ("Control+Alt+M")
SpeechOff ()
ActivateMenuBar ()
Delay (2)
TypeString ("")
EnterKey ()
TypeString ("")
EnterKey ()
PriorLine ()
Pause ()
Let nSilence = GetObjectStateCode ()
Pause ()
EnterKey ()
Pause ()
SpeechOn ()
If nSilence != 1 && nSilence !=0 Then
Say ("", ot_message)
Return
EndIf
Say (" . " + StringSegment (verb, ";", nSilence +1), ot_message)
EndScript

Script SayVoiceNone ()
If IsSameScript () Then
Let VoiceNone = ""
Else
Say (VoiceNone, ot_message)
EndIf
EndScript
